from unittest import TestCase

from mock import patch

from regulations.generator.section_url import SectionUrl


class SectionUrlTest(TestCase):
    @patch('regulations.generator.section_url.fetch_toc')
    def test_interp(self, fetch_toc):
        fetch_toc.return_value = [
            {'index': ['200', '1'], 'is_section': True},
            {'index': ['200', '2'], 'is_section': True},
            {'index': ['200', 'A'], 'is_appendix': True}]
        self.assertEqual('200-Subpart-Interp',
                         SectionUrl().interp(['200', '1', 'Interp'], 'vvvv'))
        self.assertEqual('200-Subpart-Interp',
                         SectionUrl().interp(['200', '2', 'Interp'], 'vvvv'))
        self.assertEqual('200-Appendices-Interp',
                         SectionUrl().interp(['200', 'A', 'Interp'], 'vvvv'))

        fetch_toc.return_value = [
            {'index': ['200', 'Subpart', 'A'], 'is_subpart': True,
             'sub_toc': [{'index': ['200', '1'], 'is_section': True},
                         {'index': ['200', '2'], 'is_section': True}]},
            {'index': ['200', 'A'], 'is_appendix': True},
            {'index': ['200', 'Interp'], 'is_supplement': True,
             'sub_toc': [{'index': ['200', 'Interp', 'h1'],
                          'section_id': '200-Interp-h1'},
                         {'index': ['200', 'Subpart', 'A', 'Interp'],
                          'is_subterp': True},
                         {'index': ['200', 'Appendices', 'Interp'],
                          'is_subterp': True}]}]
        self.assertEqual('200-Subpart-A-Interp',
                         SectionUrl().interp(['200', '1', 'Interp'], 'vvvv'))
        self.assertEqual('200-Subpart-A-Interp',
                         SectionUrl().interp(['200', '2', 'Interp'], 'vvvv'))
        self.assertEqual('200-Appendices-Interp',
                         SectionUrl().interp(['200', 'A', 'Interp'], 'vvvv'))
        self.assertEqual('200-Interp-h1',
                         SectionUrl().interp(['200', 'Interp', 'h1', 'p1'],
                                             'vvvv'))
        self.assertEqual('200-Interp-h1',
                         SectionUrl().interp(['200', 'Interp', 'h1'], 'vvvv'))
        self.assertTrue('200-Subpart-A',
                        SectionUrl().interp(['200', '2', 'e', 'Interp', '1'],
                                            'verver'))

    def test_of(self):
        url = SectionUrl.of(['303', '1'], 'vvv', False)
        self.assertEquals('#303-1', url)

        url = SectionUrl.of(['303', '1', 'b'], 'vvv', False)
        self.assertEquals('#303-1-b', url)

        url = SectionUrl.of(['303'], 'vvv', False)
        self.assertEquals('#303', url)

        url = SectionUrl.of(['303', '1', 'b'], 'vvv', True)
        self.assertEquals('/303-1/vvv#303-1-b', url)

        self.assertTrue('999-88/verver#999-88-e' in
                        SectionUrl.of(['999', '88', 'e'], 'verver', True))
        self.assertEqual(
            '#999-88-e', SectionUrl.of(['999', '88', 'e'], 'verver', False))

        self.assertEqual(
            '#999-Subpart-Interp',
            SectionUrl.of(['999', 'Subpart', 'Interp'], 'verver', False))
        self.assertEqual(
            '#999-Subpart-A-Interp',
            SectionUrl.of(['999', 'Subpart', 'A', 'Interp'], 'verver', False))
        self.assertEqual(
            '#999-Appendices-Interp',
            SectionUrl.of(['999', 'Appendices', 'Interp'], 'verver', False))



from django.template import loader

from regulations.generator.layers import utils
from regulations.generator.layers.base import SearchReplaceLayer


class GraphicsLayer(SearchReplaceLayer):
    shorthand = 'graphics'
    data_source = 'graphics'

    def __init__(self, layer):
        self.layer = layer
        self.template = loader.get_template('regulations/layers/graphics.html')

    def replacements_for(self, original, data):
        """Replace all instances of graphics with an img tag"""
        context = {'url': data['url'], 'alt': data['alt']}
        if 'thumb_url' in data:
            context['thumb_url'] = data['thumb_url']

        yield utils.render_template(self.template, context)



from regulations.generator.section_url import SectionUrl
from regulations.generator.toc import fetch_toc


def get_labels(current):
    return current.split('-')


def _add_extra(el, version):
    """Add extra fields to a TOC element -- only added to elements we will
    use for prev/next"""
    if el.get('is_section_span'):
        el['markup_prefix'] = '&sect;&sect;&nbsp;'
    elif el.get('is_section'):
        el['markup_prefix'] = '&sect;&nbsp;'
    elif el.get('is_subterp'):
        el['markup_prefix'] = 'Interpretations For '
    el['url'] = SectionUrl.of(el['index'], version, sectional=True)
    return el


def nav_sections(current, version):
    labels = get_labels(current)
    reg_part = labels[0]
    toc = fetch_toc(reg_part, version, flatten=True)
    for idx, el in enumerate(toc):
        if el['index'] == labels:
            if idx == 0:
                previous_section = None
            else:
                previous_section = _add_extra(toc[idx - 1], version)

            if idx == len(toc) - 1:
                next_section = None
            else:
                next_section = _add_extra(toc[idx + 1], version)

            return (previous_section, next_section)
    # Implicit return None if the section isn't in the TOC



from django.views.generic.base import TemplateView

from regulations.generator import generator
from regulations.generator.node_types import label_to_text, type_from_label
from regulations.generator.section_url import SectionUrl
from regulations.generator.sidebar.help import Help as HelpSideBar
from regulations.generator.subterp import filter_by_subterp
from regulations.generator.toc import fetch_toc
from regulations.generator.versions import fetch_grouped_history
from regulations.views import utils
from regulations.views.partial_interp import PartialSubterpView
from regulations.views.reg_landing import regulation_exists, get_versions
from regulations.views.reg_landing import regulation as landing_page
from regulations.views.partial import PartialSectionView
from regulations.views.partial_search import PartialSearch
from regulations.views.sidebar import SideBarView
from regulations.views import error_handling


class ChromeView(TemplateView):
    """ Base class for views which wish to include chrome. """
    template_name = 'regulations/chrome.html'
    #   Which view name to use when switching versions
    version_switch_view = 'chrome_section_view'
    sidebar_components = SideBarView.components
    partial_class = None

    def check_tree(self, context):
        """Throw an exception if the requested section doesn't exist"""
        label_id, version = context['label_id'], context['version']
        relevant_tree = generator.get_tree_paragraph(label_id, version)
        if relevant_tree is None:
            raise error_handling.MissingSectionException(label_id, version,
                                                         context)

    def get(self, request, *args, **kwargs):
        """Override GET so that we can catch and propagate any errors in the
        included partial(s)"""

        try:
            return super(ChromeView, self).get(request, *args, **kwargs)
        except BadComponentException as e:
            return e.response
        except error_handling.MissingSectionException as e:
            return error_handling.handle_missing_section_404(
                request, e.label_id, e.version, e.context)
        except error_handling.MissingContentException as e:
            return error_handling.handle_generic_404(request)

    def _assert_good(self, response):
        if response.status_code != 200:
            raise BadComponentException(response)

    def add_main_content(self, context):
        view = self.partial_class()
        view.request = self.request
        context['main_content_context'] = view.get_context_data(**context)
        context['main_content_template'] = view.template_name

    def diff_redirect_label(self, label_id, toc):
        """We only display diffs for sections and appendices. All other types
        of content must be converted to an appropriate diff label"""
        label_parts = label_id.split('-')
        if len(label_parts) == 1:   # whole CFR part. link to first section
            while toc:
                label_id = toc[0]['section_id']
                toc = toc[0].get('sub_toc')
        # We only show diffs for the whole interpretation at once
        elif 'Interp' in label_parts:
            label_id = label_parts[0] + '-Interp'
        # Non-section paragraph; link to the containing section
        elif len(label_parts) > 2:
            label_id = '-'.join(label_parts[:2])
        return label_id

    def set_chrome_context(self, context, reg_part, version):
        context['reg_part'] = reg_part
        context['history'] = fetch_grouped_history(reg_part)

        toc = fetch_toc(reg_part, version)
        for el in toc:
            el['url'] = SectionUrl().of(
                el['index'], version, self.partial_class.sectional_links)
            for sub in el.get('sub_toc', []):
                sub['url'] = SectionUrl().of(
                    sub['index'], version, self.partial_class.sectional_links)
        context['TOC'] = toc

        context['meta'] = utils.regulation_meta(reg_part, version)
        context['version_switch_view'] = self.version_switch_view
        context['diff_redirect_label'] = self.diff_redirect_label(
            context['label_id'], toc)

    def get_context_data(self, **kwargs):
        context = super(ChromeView, self).get_context_data(**kwargs)

        label_id = context['label_id']
        version = context['version']
        label_id_list = label_id.split('-')
        reg_part = label_id_list[0]
        context['q'] = self.request.GET.get('q', '')
        context['formatted_id'] = label_to_text(label_id_list, True, True)
        context['node_type'] = type_from_label(label_id_list)

        error_handling.check_regulation(reg_part)
        self.set_chrome_context(context, reg_part, version)

        self.check_tree(context)
        self.add_main_content(context)
        context['sidebar_content'] = self.sidebar(label_id, version)

        return context

    def sidebar(self, label_id, version):
        """Generate the sidebar content for this label_id+version. This
        involves passing through to the SideBarView"""
        sidebar_view = SideBarView.as_view(components=self.sidebar_components)
        response = sidebar_view(self.request, label_id=label_id,
                                version=version)
        self._assert_good(response)
        response.render()
        return response.content


class ChromeSubterpView(ChromeView):
    """Corresponding chrome class for subterp partial view"""
    partial_class = PartialSubterpView
    version_switch_view = 'chrome_subterp_view'

    def check_tree(self, context):
        """We can't defer to Chrome's check because Subterps are constructed
        -site side"""
        version, label_id = context['version'], context['label_id']
        label = label_id.split('-')
        reg_part = label[0]

        interp = generator.get_tree_paragraph(reg_part + '-Interp', version)
        if not interp:
            raise error_handling.MissingSectionException(label_id, version,
                                                         context)

        subterp_sects = filter_by_subterp(interp['children'], label, version)
        if not subterp_sects:
            raise error_handling.MissingSectionException(label_id, version,
                                                         context)


class ChromeSearchView(ChromeView):
    """Search results with chrome"""
    template_name = 'regulations/chrome-search.html'
    partial_class = PartialSearch
    sidebar_components = [HelpSideBar]

    def check_tree(self, context):
        pass    # Search doesn't perform this check

    def get_context_data(self, **kwargs):
        """Get the version for the chrome context"""
        kwargs['version'] = self.request.GET.get('version', '')
        kwargs['skip_count'] = True
        kwargs['label_id'] = utils.first_section(kwargs['label_id'],
                                                 kwargs['version'])
        return super(ChromeSearchView, self).get_context_data(**kwargs)

    def add_main_content(self, context):
        """Override this so that we have access to the main content's
        results field"""
        super(ChromeSearchView, self).add_main_content(context)
        context['results'] = context['main_content_context']['results']


class ChromeLandingView(ChromeView):
    """Landing page with chrome"""
    template_name = 'regulations/landing-chrome.html'
    partial_class = PartialSectionView  # Needed to know sectional status

    def check_tree(self, context):
        pass    # Landing page doesn't perform this check

    def sidebar(self, label_id, version):
        """Landing pages don't have a sidebar generated this way"""
        return None

    def add_main_content(self, context):
        """Landing page isn't a TemplateView"""
        response = landing_page(self.request, context['reg_part'])
        self._assert_good(response)
        context['main_content'] = response.content

    def get_context_data(self, **kwargs):
        """Add the version and replace the label_id for the chrome context"""

        reg_part = kwargs['label_id']
        if not regulation_exists(reg_part):
            raise error_handling.MissingContentException()

        current, _ = get_versions(kwargs['label_id'])
        kwargs['version'] = current['version']
        kwargs['label_id'] = utils.first_section(reg_part, current['version'])
        return super(ChromeLandingView, self).get_context_data(**kwargs)


class BadComponentException(Exception):
    """Allows us to propagate errors in loaded partials"""
    def __init__(self, response):
        self.response = response

    def __str__(self):
        return repr(self)

    def __repr__(self):
        return "BadComponentException(response=%s)" % repr(self.response)



# -*- coding: utf-8 -*-
import unittest

from selenium.webdriver.support.ui import WebDriverWait

from regulations.uitests.base_test import BaseTest


class NavigationTest(BaseTest, unittest.TestCase):

    job_name = 'Navigation test'

    def test_navigation(self):
        self.driver.get(self.test_url + '/1005-5/2012-12121')
        html = self.driver.find_element_by_tag_name('html')
        WebDriverWait(self.driver, 30).until(
            lambda driver: 'selenium-start' in html.get_attribute('class'))

        self.driver.execute_script(
            'poffset = document.getElementById("1005-5-b-2").offsetTop')
        self.driver.execute_script('window.scrollTo(0, poffset)')

        # wayfinding header should update
        self.driver.implicitly_wait(5)
        header = self.driver.find_element_by_css_selector('#active-title')
        self.assertIn(header.text,
                      (u'§ 1005.5(b)', u'§ 1005.5(b)(1)', u'§ 1005.5(b)(2)'))

        fwd_link = self.driver.find_element_by_css_selector('li.next a')
        fwd_link.click()
        # should navigate to next section
        WebDriverWait(self.driver, 30).until(
            lambda driver: driver.find_element_by_id('1005-6'))

        self.driver.back()

        # should go back to 1005-5
        self.driver.find_element_by_xpath('//*[@id="1005-5"]')
        back_link = self.driver.find_element_by_css_selector('li.previous a')
        back_link.click()
        # should navigate to 1005-4
        WebDriverWait(self.driver, 30).until(
            lambda driver: driver.find_element_by_id('1005-4'))



import platform
import os.path
import subprocess   # nosec - see usage below

from django.conf import settings
from django.core.management.base import BaseCommand, CommandError


BIN_PATH = (
    'http://download.gna.org', 'wkhtmltopdf', '0.12', '0.12.3',
    'wkhtmltox-0.12.3_linux-generic-amd64.tar.xz',
)


class Command(BaseCommand):
    help = 'Fetch wkhtmltox binary'

    def add_arguments(self, parser):
        parser.add_argument(
            '--force', action='store_true', dest='force', default=False,
            help="Download the binary even if it's present")

    def handle(self, **options):
        if platform.system() != 'Linux':
            raise CommandError(
                'The `wkhtmltox` command only handles linux; to install on '
                'another platform, see http://wkhtmltopdf.org/downloads.html.'
            )

        path = settings.WKHTMLTOPDF_PATH
        if path and os.path.exists(path) and not options['force']:
            message = ("WKHTMLTOPDF already exists: {}\n"
                       "Skipping. Use the --force flag if necessary.")
            self.stderr.write(self.style.NOTICE(
                message.format(settings.WKHTMLTOPDF_PATH)))
        else:
            # Safe because: we're not passing user input into these processes
            subprocess.check_call([     # nosec
                'wget', '-O', BIN_PATH[-1], '/'.join(BIN_PATH)])
            subprocess.check_call(['tar', 'xvf', BIN_PATH[-1]])     # nosec



def get_notice(api_client, document_number):
    return api_client.notice(document_number)


def filter_labeled_children(sxs):
    """ Some children don't have labels. We display those with their parents.
    The other children are displayed when they are independently, specifically
    requested. """
    return [s for s in sxs['children'] if 'label' not in s]


def non_empty_sxs(sxs):
    has_paragraphs = len(sxs['paragraphs']) > 0
    has_unlabeled_children = len(filter_labeled_children(sxs)) > 0
    return (has_paragraphs or has_unlabeled_children)


def add_depths(sxs, starting_depth):
    """ We use depth numbers in header tags  to determine how titles are
    output. """

    sxs['depth'] = starting_depth
    for s in sxs['children']:
        add_depths(s, starting_depth+1)


def find_label_in_sxs(sxs_list, label_id, fr_page=None):
    """ Given a tree of SXS sections, find a non-empty sxs that matches
    label_id. Some notices may have the same label appearing multiple times;
    use fr_page to distinguish, defaulting to the first"""

    matches = []

    for s in sxs_list:
        if label_id in s.get('labels', [s.get('label')]) and non_empty_sxs(s):
            matches.append(s)
        elif s['children']:
            sxs = find_label_in_sxs(s['children'], label_id, fr_page)
            if sxs and non_empty_sxs(sxs):
                matches.append(sxs)

    perfect_match = [m for m in matches if m.get('page') == fr_page]
    if perfect_match:
        return perfect_match[0]
    if matches:
        return matches[0]



from unittest import TestCase
from mock import patch

from regulations.views import reg_landing


class LandingViewTest(TestCase):
    @patch('regulations.views.reg_landing.api_reader')
    def test_regulation_exists_not(self, api_reader):
        api_reader.ApiReader.return_value.regversions.return_value = None
        reg_versions = reg_landing.regulation_exists('204')
        self.assertFalse(reg_versions)

    @patch('regulations.views.reg_landing.api_reader')
    def test_regulation_exists(self, api_reader):
        api_reader.ApiReader.return_value.regversions.return_value = [
            {'version': 'exists'}]
        reg_versions = reg_landing.regulation_exists('204')
        self.assertTrue(reg_versions)

    @patch('regulations.views.reg_landing.fetch_grouped_history')
    def test_get_versions(self, fetch_grouped_history):
        fetch_grouped_history.return_value = [
            {'timeline': 'future', 'version': 'a'},
            {'timeline': 'current', 'version': 'b'}]
        current_ver, next_ver = reg_landing.get_versions('204')
        self.assertEqual({'timeline': 'current', 'version': 'b'}, current_ver)
        self.assertEqual({'timeline': 'future', 'version': 'a'}, next_ver)

    @patch('regulations.views.reg_landing.fetch_grouped_history')
    def test_get_versions_no_next(self, fetch_grouped_history):
        fetch_grouped_history.return_value = [
            {'timeline': 'current', 'version': 'b'}]
        current_ver, next_ver = reg_landing.get_versions('204')
        self.assertEqual({'timeline': 'current', 'version': 'b'}, current_ver)
        self.assertEqual(None, next_ver)



from importlib import import_module
import logging
import re
from threading import Thread

from django.conf import settings

from regulations.generator import api_reader
from regulations.generator.layers.base import LayerBase
from regulations.generator.layers.layers_applier import (
    InlineLayersApplier, ParagraphLayersApplier, SearchReplaceLayersApplier)
from regulations.generator.layers.diff_applier import DiffApplier
from regulations.generator import notices


def _data_layers():
    """Index all configured data layers by their "shorthand". This doesn't
    have any error checking -- it'll explode if configured improperly"""
    layers = {}
    for class_path in settings.DATA_LAYERS:
        module, class_name = class_path.rsplit('.', 1)
        klass = getattr(import_module(module), class_name)
        layers[klass.shorthand] = klass
    return layers


class LayerCreator(object):
    """ This lets us dynamically load layers by shorthand. """
    LAYERS = _data_layers()

    def __init__(self):
        self.appliers = {
            LayerBase.INLINE: InlineLayersApplier(),
            LayerBase.PARAGRAPH: ParagraphLayersApplier(),
            LayerBase.SEARCH_REPLACE: SearchReplaceLayersApplier()}

        self.api = api_reader.ApiReader()

    def get_layer_json(self, layer_name, doc_type, label_id, version=None):
        """Hit the API to retrieve layer data"""
        return self.api.layer(layer_name, doc_type, label_id, version)

    def add_layers(self, layer_names, doc_type, label_id, sectional=False,
                   version=None):
        """Request a list of layers. As this might spawn multiple HTTP
        requests, we wrap the requests in threads so they can proceed
        concurrently."""
        # This doesn't deal with sectional interpretations yet.
        # we'll have to do that.
        layer_names = set(l for l in layer_names
                          if l.lower() in LayerCreator.LAYERS)
        results = []
        procs = []

        def one_layer(layer_name):
            layer_class = LayerCreator.LAYERS[layer_name]
            api_name = layer_class.data_source
            applier_type = layer_class.layer_type
            layer_json = self.get_layer_json(api_name, doc_type, label_id,
                                             version)
            results.append((api_name, applier_type, layer_class, layer_json))

        #   Spawn threads
        for layer_name in layer_names:
            proc = Thread(target=one_layer, args=(layer_name,))
            procs.append(proc)
            proc.start()

        #   Join them (once their work is done)
        for proc in procs:
            proc.join()

        for api_name, applier_type, layer_class, layer_json in results:
            if layer_json is None:
                logging.warning("No data for %s %s %s %s", api_name, doc_type,
                                label_id, version)
            else:
                layer = layer_class(layer_json)

                if sectional and hasattr(layer, 'sectional'):
                    layer.sectional = sectional
                if hasattr(layer, 'version'):
                    layer.version = version

                self.appliers[applier_type].add_layer(layer)

    def get_appliers(self):
        """ Return the appliers. """
        return (self.appliers[LayerBase.INLINE],
                self.appliers[LayerBase.PARAGRAPH],
                self.appliers[LayerBase.SEARCH_REPLACE])


class DiffLayerCreator(LayerCreator):
    def __init__(self, newer_version):
        super(DiffLayerCreator, self).__init__()
        self.newer_version = newer_version

    def get_layer_json(self, layer_name, doc_type, label_id, version):
        """Diffs contain layer data from _two_ documents, each corresponding
        to one of the versions we're comparing. This data is then combined
        before displaying"""
        older_layer = self.api.layer(layer_name, doc_type, label_id, version)
        newer_layer = self.api.layer(layer_name, doc_type, label_id,
                                     self.newer_version)

        layer_json = dict(newer_layer)  # copy
        layer_json.update(older_layer)  # older layer takes precedence
        return layer_json


def get_regulation(regulation, version):
    """ Get the regulation JSON tree. Manipulate the label a bit for easier
    access in the templates."""
    api = api_reader.ApiReader()
    reg = api.regulation(regulation, version)

    if reg:
        title = reg['title']
        # up till the paren
        match = re.search('part \d+[^\w]*([^\(]*)', title, re.I)
        if match:
            reg['title_clean'] = match.group(1).strip()
        match = re.search('\(regulation (\w+)\)', title, re.I)
        if match:
            reg['reg_letter'] = match.group(1)

        return reg


def get_tree_paragraph(paragraph_id, version):
    """Get a single level of the regulation tree."""
    api = api_reader.ApiReader()
    return api.regulation(paragraph_id, version)


def get_notice(document_number):
    """ Get a the data from a particular notice, given the Federal Register
    document number. """

    api = api_reader.ApiReader()
    return api.notice(document_number)


def get_sxs(label_id, notice, fr_page=None):
    """ Given a paragraph label_id, find the sxs analysis for that paragraph if
    it exists and has content. fr_page is used to distinguish between
    multiple analyses in the same notice."""

    all_sxs = notice['section_by_section']
    relevant_sxs = notices.find_label_in_sxs(all_sxs, label_id, fr_page)

    return relevant_sxs


def get_diff_json(regulation, older, newer):
    api = api_reader.ApiReader()
    return api.diff(regulation, older, newer)


def get_diff_applier(label_id, older, newer):
    regulation = label_id.split('-')[0]
    diff_json = get_diff_json(regulation, older, newer)
    if diff_json is not None:
        return DiffApplier(diff_json, label_id)



from django.conf.urls import url

from regulations.url_caches import daily_cache, lt_cache
from regulations.views.about import about
from regulations.views.chrome_breakaway import ChromeSXSView
from regulations.views.chrome import (
    ChromeView, ChromeLandingView,
    ChromeSearchView,
    ChromeSubterpView)
from regulations.views.diff import ChromeSectionDiffView
from regulations.views.diff import PartialSectionDiffView
from regulations.views.partial import PartialDefinitionView
from regulations.views.partial import PartialParagraphView
from regulations.views.partial import PartialRegulationView, PartialSectionView
from regulations.views import partial_interp
from regulations.views.partial_search import PartialSearch
from regulations.views.partial_sxs import ParagraphSXSView
from regulations.views.preamble import (
    CFRChangesView, PreambleView, PrepareCommentView,
    ChromePreambleSearchView)
from regulations.views.redirect import diff_redirect, redirect_by_date
from regulations.views.redirect import redirect_by_date_get
from regulations.views.sidebar import SideBarView
from regulations.views.universal_landing import universal
from regulations.views import comment

# Re-usable URL patterns.
meta_version = r'(?P<%s>[-\d\w_]+)'
version_pattern = meta_version % 'version'
newer_version_pattern = meta_version % 'newer_version'
notice_pattern = meta_version % 'notice_id'

reg_pattern = r'(?P<label_id>[\d]+)'
preamble_pattern = r'(?P<label_id>[\w]+)'
section_pattern = r'(?P<label_id>[\d]+[-][\w]+)'
interp_pattern = r'(?P<label_id>[-\w]+[-]Interp)'
paragraph_pattern = r'(?P<label_id>[-\w]+)'
subterp_pattern = r'(?P<label_id>[\d]+-(Appendices|Subpart(-[A-Z]+)?)-Interp)'


urlpatterns = [
    url(r'^$', universal, name='universal_landing'),
    # about page
    url(r'^about$', about, name='about'),
    url(r'^comments/attachment$', comment.upload_proxy),
    url(r'^comments/review/(?P<doc_number>[\w-]+)$',
        PrepareCommentView.as_view(), name='comment_review'),
    url(r'^comments/preview$', comment.preview_comment),
    url(r'^comments/comment/(?P<doc_number>[\w-]+)$',
        comment.SubmitCommentView.as_view(),
        name='comment_submit'),
    url(r'^comments/federal_agencies$', comment.get_federal_agencies),
    url(r'^comments/gov_agency_types$', comment.get_gov_agency_types),
    # Redirect to version by date (by GET)
    # Example http://.../regulation_redirect/201-3-v
    url(r'^regulation_redirect/%s$' % paragraph_pattern, redirect_by_date_get,
        name='redirect_by_date_get'),
    # Redirect to a diff based on GET params
    # Example http://.../diff_redirect/201-3/old_version?new_version=new
    url(r'^diff_redirect/%s/%s$' % (section_pattern, version_pattern),
        diff_redirect, name='diff_redirect'),
    # A section by section paragraph with chrome
    # Example: http://.../sxs/201-2-g/2011-1738
    url(r'^sxs/%s/%s$' % (paragraph_pattern, notice_pattern),
        lt_cache(ChromeSXSView.as_view()),
        name='chrome_sxs_view'),
    # Search results for non-JS viewers
    # Example: http://.../search?q=term&version=2011-1738
    url(r'^search(?:/cfr)?/%s$' % reg_pattern,
        ChromeSearchView.as_view(), name='chrome_search',
        kwargs={'doc_type': 'cfr'}),
    url(r'^search/preamble/%s$' % preamble_pattern,
        ChromePreambleSearchView.as_view(), name='chrome_search_preamble',
        kwargs={'doc_type': 'preamble'}),
    # Diff view of a section for non-JS viewers (or book markers)
    # Example: http://.../diff/201-4/2011-1738/2013-10704
    url(r'^diff/%s/%s/%s$' %
        (section_pattern, version_pattern, newer_version_pattern),
        lt_cache(ChromeSectionDiffView.as_view()),
        name='chrome_section_diff_view'),

    url(r'^preamble/(?P<doc_number>[\w-]+)/cfr_changes/(?P<section>[\w-]+)$',
        daily_cache(CFRChangesView.as_view()), name='cfr_changes'),
    url(r'^preamble/(?P<paragraphs>[-\w]+(/[-\w]+)*)$',
        daily_cache(PreambleView.as_view()), name='chrome_preamble'),

    # Redirect to version by date
    # Example: http://.../201-3-v/1999/11/8
    url(r'^%s/(?P<year>\d{4})/(?P<month>\d{2})/(?P<day>\d{2})$'
        % paragraph_pattern, redirect_by_date, name='redirect_by_date'),
    # A regulation section with chrome
    # Example: http://.../201-4/2013-10704
    url(r'^%s/%s$' % (section_pattern, version_pattern),
        lt_cache(ChromeView.as_view(partial_class=PartialSectionView)),
        name='chrome_section_view'),
    # Subterp, interpretations of a while subpart, emptypart or appendices
    # Example: http://.../201-Subpart-A-Interp/2013-10706
    #          http://.../201-Subpart-Interp/2013-10706
    #          http://.../201-Appendices-Interp/2013-10706
    url(r'^%s/%s$' % (subterp_pattern, version_pattern),
        lt_cache(ChromeSubterpView.as_view()),
        name=ChromeSubterpView.version_switch_view),
    # Interpretation of a section/paragraph or appendix
    # Example: http://.../201-4-Interp/2013-10704
    url(r'^%s/%s$' % (interp_pattern, version_pattern),
        lt_cache(ChromeView.as_view(
            partial_class=partial_interp.PartialInterpView)),
        name='chrome_interp_view'),
    # The whole regulation with chrome
    # Example: http://.../201/2013-10704
    url(r'^%s/%s$' % (reg_pattern, version_pattern),
        lt_cache(ChromeView.as_view(
            partial_class=PartialRegulationView,
            version_switch_view='chrome_regulation_view')),
        name='chrome_regulation_view'),
    # A regulation paragraph with chrome
    # Example: http://.../201-2-g/2013-10704
    url(r'^%s/%s$' % (paragraph_pattern, version_pattern),
        lt_cache(ChromeView.as_view(
            partial_class=PartialParagraphView,
            version_switch_view='chrome_paragraph_view')),
        name='chrome_paragraph_view'),
    # A regulation landing page
    # Example: http://.../201
    url(r'^%s$' % reg_pattern, ChromeLandingView.as_view(),
        name='regulation_landing_view'),

    # Load just the sidebar
    # Example: http://.../partial/sidebar/201-2/2013-10704
    url(r'^partial/sidebar/%s/%s$' % (paragraph_pattern, version_pattern),
        SideBarView.as_view(),
        name='sidebar'),

    # Load just search results
    url(r'^partial/search(?:/cfr)?/%s$' % reg_pattern,
        PartialSearch.as_view(), name='partial_search',
        kwargs={'doc_type': 'cfr'}),
    url(r'^partial/search/preamble/%s$' % preamble_pattern,
        PartialSearch.as_view(), name='partial_search',
        kwargs={'doc_type': 'preamble'}),

    # A diff view of a section (without chrome)
    url(r'^partial/diff/%s/%s/%s$' % (
        section_pattern, version_pattern, newer_version_pattern),
        lt_cache(PartialSectionDiffView.as_view()),
        name='partial_section_diff_view'),
    # A section by section paragraph (without chrome)
    # Example: http://.../partial/sxs/201-2-g/2011-1738
    url(r'^partial/sxs/%s/%s$' % (paragraph_pattern, notice_pattern),
        lt_cache(ParagraphSXSView.as_view()),
        name='paragraph_sxs_view'),
    # A definition templated to be displayed in the sidebar (without chrome)
    # Example: http://.../partial/definition/201-2-g/2011-1738
    url(r'^partial/definition/%s/%s$' % (paragraph_pattern, version_pattern),
        lt_cache(PartialDefinitionView.as_view()),
        name='partial_definition_view'),
    # A regulation section without chrome
    # Example: http://.../partial/201-4/2013-10704
    url(r'^partial/%s/%s$' % (section_pattern, version_pattern),
        lt_cache(PartialSectionView.as_view()),
        name='partial_section_view'),
    # Subterp, interpretations of a whole subpart, emptypart or appendices
    # Example: http://.../partial/201-Subpart-A-Interp/2013-10706
    #          http://.../partial/201-Subpart-Interp/2013-10706
    #          http://.../partial/201-Appendices-Interp/2013-10706
    url(r'^partial/%s/%s$' % (subterp_pattern, version_pattern),
        lt_cache(partial_interp.PartialSubterpView.as_view()),
        name='partial_subterp_view'),
    # An interpretation of a section/paragraph or appendix without chrome.
    # Example: http://.../partial/201-2-Interp/2013-10704
    url(r'^partial/%s/%s$' % (interp_pattern, version_pattern),
        lt_cache(partial_interp.PartialInterpView.as_view()),
        name='partial_interp_view'),
    # The whole regulation without chrome; not too useful; added for symmetry
    # Example: http://.../partial/201/2013-10704
    url(r'^partial/%s/%s$' % (reg_pattern, version_pattern),
        lt_cache(PartialRegulationView.as_view()),
        name='partial_regulation_view'),
    # A regulation paragraph without chrome.
    # Example: http://.../partial/201-2-g/2013-10704
    url(r'^partial/%s/%s$' % (paragraph_pattern, version_pattern),
        lt_cache(PartialParagraphView.as_view()),
        name='partial_paragraph_view'),
]



from django.http import Http404

from regulations.generator import generator, node_types
from regulations.generator.subterp import filter_by_subterp
from regulations.views.partial import (
    PartialSectionView, PartialView, generate_html)


class PartialInterpView(PartialView):
    """ Interpretation of a reg text section/paragraph or appendix. Used for
    in-line interpretations"""

    template_name = "regulations/interpretations.html"
    inline = False
    appliers = []

    @staticmethod
    def mk_appliers(root_label, version):
        """Function to generate a shared set of appliers"""
        layer_creator = generator.LayerCreator()
        layer_creator.add_layers(
            ['terms', 'internal', 'keyterms', 'paragraph'], 'cfr', root_label,
            sectional=True, version=version)
        return layer_creator.get_appliers()

    def determine_appliers(self, label_id, version):
        """Don't generate new appliers"""
        return self.appliers

    def transform_context(self, context, builder):
        context['inline'] = self.inline
        context['c'] = {'node_type': node_types.INTERP,
                        'children': [builder.tree]}
        return context


class PartialSubterpView(PartialSectionView):
    """View of subterps - interpretations of whole subparts, regtext, or
    appendices"""
    def get_context_data(self, **kwargs):
        #   skip our parent
        context = super(PartialView, self).get_context_data(**kwargs)

        label_id = context['label_id']
        label = label_id.split('-')
        reg_part = label[0]
        version = context['version']

        context['navigation'] = self.section_navigation(label_id, version)

        interp = generator.get_tree_paragraph(reg_part + '-Interp', version)
        if not interp:
            raise Http404

        subterp_sects = filter_by_subterp(interp['children'], label, version)
        if not subterp_sects:
            raise Http404

        context['markup_page_type'] = 'reg-section'
        html_label = node_types.to_markup_id(label_id.split('-'))
        interp['children'] = subterp_sects

        # interp['label] is defined so that the template receives the
        # appropriate markup ID, matching the rendered subterp and not
        # the parent node in the tree
        interp['label'] = label
        inline_applier, p_applier, s_applier = self.determine_appliers(
            reg_part + '-Interp', version)
        builder = generate_html(interp, (inline_applier, p_applier, s_applier))
        interp = builder.tree
        interp['html_label'] = html_label
        context['tree'] = {'children': [interp]}
        return context



# vim: set fileencoding=utf-8
import re

from regulations.generator import node_types

# A section title comprises
#   - one or more §
#   - a section label like 11.45 or 11.45-50
#   - one or more space or - separators
#   - a section subject
SECTION_TITLE_REGEX = re.compile(u'^§+ ([-.\w]*)[\s-]*(.*)', re.UNICODE)


def appendix_supplement(data):
    """Handle items pointing to an appendix or supplement"""
    node_type = node_types.type_from_label(data['index'])
    if len(data['index']) == 2 and node_type in (node_types.APPENDIX,
                                                 node_types.INTERP):
        element = {}
        if node_type == node_types.INTERP:
            element['is_supplement'] = True
        else:
            element['is_appendix'] = True

        segments = try_split(data['title'])
        if segments:
            element['label'], element['sub_label'] = segments[:2]
        elif '[' in data['title']:
            position = data['title'].find('[')
            element['label'] = data['title'][:position].strip()
            element['sub_label'] = data['title'][position:]
        else:
            element['label'] = data['title']

        element['section_id'] = '-'.join(data['index'])
        return element


def try_split(text, chars=(u'—', '-')):
    """Utility method for splitting a string by one of multiple chars"""
    for c in chars:
        segments = text.split(c)
        if len(segments) > 1:
            return [s.strip() for s in segments]


def section(data):
    """ Parse out parts of a section title. """
    if len(data['index']) == 2 and data['index'][1][0].isdigit():
        element = {}
        element['is_section'] = True
        element['section_id'] = '-'.join(data['index'])
        if u"§§ " == data['title'][:3]:
            element['is_section_span'] = True
        else:
            element['is_section_span'] = False
        match = SECTION_TITLE_REGEX.match(data['title'])
        element['label'] = match.group(1)
        element['sub_label'] = match.group(2)
        return element



# vim: set fileencoding=utf-8
from unittest import TestCase

from django.test import RequestFactory
from mock import patch

from regulations.views import diff as views_diff


class ChromeSectionDiffViewTests(TestCase):

    def test_extract_sections_subparts(self):
        sub1 = [{'section_id': '8888-1', 'index': ['8888', '1']},
                {'section_id': '8888-3', 'index': ['8888', '3']}]
        sub2 = [{'section_id': '8888-7', 'index': ['8888', '7']}]

        toc = [
            {'section_id': '8888-Subpart-A',
             'index': ['8888', 'Subpart', 'A'],
             'sub_toc': sub1},
            {'section_id': '8888-Subjgrp-IND',
             'index': ['8888', 'Subjgrp', 'IND'],
             'sub_toc': sub2},
            {'section_id': '8888-Interp', 'index': ['8888', 'Interp']},
        ]

        sections = views_diff.extract_sections(toc)
        self.assertEqual(['8888', '1'], sections[0]['index'])
        self.assertEqual(['8888', '3'], sections[1]['index'])
        self.assertEqual(['8888', '7'], sections[2]['index'])
        self.assertEqual(['8888', 'Interp'], sections[3]['index'])

    def test_extract_sections(self):
        toc = [{'section_id': '8888-1', 'index': ['8888', '1']},
               {'section_id': '8888-3', 'index': ['8888', '3']}]
        sections = views_diff.extract_sections(toc)

        self.assertEqual(['8888', '1'], sections[0]['index'])
        self.assertEqual(['8888', '3'], sections[1]['index'])

    def test_diff_toc(self):
        """Integration test."""
        versions = views_diff.Versions('oldold', 'newnew', 'from_ver')
        old_toc = [{'section_id': '8888-1', 'index': ['8888', '1'],
                    'is_section':True},
                   {'section_id': '8888-3', 'index': ['8888', '3'],
                    'is_section':True},
                   {'section_id': '8888-4', 'index': ['8888', '4'],
                    'is_section':True},
                   {'section_id': '8888-A', 'index': ['8888', 'A'],
                    'is_appendix':True},
                   {'section_id': '8888-B', 'index': ['8888', 'B'],
                    'is_appendix':True},
                   {'section_id': '8888-Interp', 'index': ['8888', 'Interp'],
                    'is_supplement':True}]
        diff = {
            '8888-2': {'op': 'added',
                       'node': {'title': u'§ 8888.2', 'label': ['8888', '2']}},
            '8888-C': {'op': 'added',
                       'node': {'title': 'App C', 'label': ['8888', 'C']}},
            '8888-1-a': {'op': 'modified'},
            '8888-B': {'op': 'deleted'},
            '8888-3-b': {'op': 'deleted'},
            '8888-B-1': {'op': 'modified'}
        }

        result = views_diff.diff_toc(versions, old_toc, diff)
        self.assertEqual(8, len(result))
        self.assertTrue('8888-1' in result[0]['url'])
        self.assertTrue('?from_version=from_ver' in result[0]['url'])
        self.assertEqual('8888-1', result[0]['section_id'])
        self.assertEqual('modified', result[0]['op'])
        self.assertTrue('8888-2' in result[1]['url'])
        self.assertTrue('?from_version=from_ver' in result[1]['url'])
        self.assertEqual('8888-2', result[1]['section_id'])
        self.assertEqual('added', result[1]['op'])
        self.assertTrue('8888-3' in result[2]['url'])
        self.assertTrue('?from_version=from_ver' in result[2]['url'])
        self.assertEqual('8888-3', result[2]['section_id'])
        self.assertEqual('modified', result[2]['op'])
        self.assertTrue('8888-4' in result[3]['url'])
        self.assertTrue('?from_version=from_ver' in result[3]['url'])
        self.assertEqual('8888-4', result[3]['section_id'])
        self.assertFalse('op' in result[3])
        self.assertTrue('8888-A' in result[4]['url'])
        self.assertTrue('?from_version=from_ver' in result[4]['url'])
        self.assertEqual('8888-A', result[4]['section_id'])
        self.assertFalse('op' in result[4])
        self.assertTrue('8888-B' in result[5]['url'])
        self.assertTrue('?from_version=from_ver' in result[5]['url'])
        self.assertEqual('8888-B', result[5]['section_id'])
        self.assertEqual('deleted', result[5]['op'])
        self.assertTrue('8888-C' in result[6]['url'])
        self.assertTrue('?from_version=from_ver' in result[6]['url'])
        self.assertEqual('8888-C', result[6]['section_id'])
        self.assertEqual('added', result[6]['op'])
        self.assertTrue('8888-Interp' in result[7]['url'])
        self.assertEqual('8888-Interp', result[7]['section_id'])
        self.assertFalse('op' in result[7])
        for el in result:
            self.assertTrue('oldold', el['url'])
            self.assertTrue('newnew', el['url'])

    def test_interp_headers(self):
        from django.template import loader, Context
        t = loader.get_template('regulations/interp-tree.html')
        context_dict = {'interp': {
            'header': '<ins>My header</ins>', 'section_header': True}}
        response = t.render(Context(context_dict))
        tags_preserved_header = '<h3 tabindex=\"0\"> <ins>My header</ins></h3>'
        self.assertTrue(tags_preserved_header in response)

    @patch('regulations.views.diff.fetch_toc')
    @patch('regulations.views.diff.SectionUrl')
    def test_add_main_content(self, SectionUrl, fetch_toc):
        fetch_toc.return_value = [
            {'index': ['1111', '222'], 'is_section': True},
            {'index': ['1111', 'B'], 'is_appendix': True},
            {'index': ['1111', 'Interp'], 'is_supplement': True,
             'sub_toc': [
                 {'index': ['1111', 'Interp', 'h1']},
                 {'index': ['1111', 'Subpart', 'Interp']},
                 {'index': ['1111', 'A', 'Interp']}]}]
        context = {
            'main_content_context': {'newer_version': '1', 'TOC': 'toc'},
            'label_id': '111-222',
            'version': '2'}
        request = RequestFactory().get('?new_version=1')
        csdv = views_diff.ChromeSectionDiffView()
        csdv.request = request

        csdv.add_diff_content(context)
        self.assertEqual(context['from_version'], '2')
        self.assertEqual(context['left_version'], '2')
        self.assertEqual(context['right_version'], '1')
        self.assertEqual(context['TOC'], 'toc')
        self.assertTrue('first_subterp' in context)
        self.assertEqual(['1111', 'Interp', 'h1'],
                         context['first_subterp']['index'])
        self.assertTrue('url' in context['first_subterp'])


class PartialSectionDiffViewTests(TestCase):
    def setUp(self):
        self.versions = views_diff.Versions('old', 'new', 'from')
        self.toc = [{'section_id': '9898-1'}, {'section_id': '9898-5'},
                    {'section_id': '9898-A'}, {'section_id': '9898-Interp'}]
        self.view = views_diff.PartialSectionDiffView()

    def assert_url_contains_versions(self, url, section_id):
        """The provided url should contain the appropriate strings for the
        old, new, and return_to urls"""
        from_version = '?from_version=' + self.versions.return_to
        self.assertTrue(self.versions.older in url)
        self.assertTrue(self.versions.newer in url)
        self.assertTrue(from_version in url)
        self.assertTrue(section_id in url)

    def assert_correct_nav(self, section_id, prev, following):
        """Verify that the generated nav contains the appropriate url entries
        for prev and next"""
        nav = self.view.footer_nav(section_id, self.toc, self.versions)
        self.assertEqual(nav['page_type'], 'diff')
        if prev:
            self.assert_url_contains_versions(nav['previous']['url'], prev)
        else:
            self.assertFalse('previous' in nav)

        if following:
            self.assert_url_contains_versions(nav['next']['url'], following)
        else:
            self.assertFalse('next' in nav)

    def test_footer_nav(self):
        self.assert_correct_nav('9898-2', prev=None, following=None)
        self.assert_correct_nav('9898-1', prev=None, following='9898-5')
        self.assert_correct_nav('9898-5', prev='9898-1', following='9898-A')
        self.assert_correct_nav('9898-A', prev='9898-5',
                                following='9898-Interp')
        self.assert_correct_nav('9898-Interp', prev='9898-A', following=None)



import unittest

from nose.tools import assert_equal, assert_in
from selenium.webdriver.support.ui import WebDriverWait

from regulations.uitests.base_test import BaseTest


class CommentTest(BaseTest, unittest.TestCase):

    job_name = 'Comment test'

    def test_comment(self):
        self.driver.get(self.test_url + '/preamble/2016_02749/I')
        html = self.driver.find_element_by_tag_name('html')
        WebDriverWait(self.driver, 60).until(
            lambda driver: 'selenium-start' in html.get_attribute('class'))

        # Open write mode for a section
        comment_toggle = self.driver.find_element_by_css_selector(
            '.activate-write')
        comment_label = comment_toggle.get_attribute('data-label')
        comment_toggle.click()

        # Comment index starts empty
        item_container = self.driver.find_element_by_css_selector(
            '.comment-index-items')
        self.driver.implicitly_wait(0)
        index_items = item_container.find_elements_by_css_selector(
            '.comment-index-item')
        assert_equal(len(index_items), 0)
        self.driver.implicitly_wait(30)

        # Write and save a comment
        textarea = self.driver.find_element_by_css_selector(
            '.ProseMirror-content')
        textarea.send_keys('i prefer not to')
        self.driver.find_element_by_css_selector(
            '.comment button[type="submit"]').click()

        # Comment index is populated
        index_items = self.driver.find_elements_by_css_selector(
            '.comment-index-item')
        assert_equal(len(index_items), 1)
        assert_in(comment_label, index_items[0].text)

        # Browse to review page
        self.driver.execute_script('window.scrollTo(0, 0);')
        self.driver.find_element_by_css_selector(
            '.comment-index-review').click()

        # Verify comment in review text
        html = self.driver.find_element_by_css_selector(
            '.comments').get_attribute('innerHTML')
        assert_in(comment_label, html)
        assert_in('i prefer not to', html)

    def test_intro(self):
        """Verify that the intro meta data is visible"""
        self.driver.get(self.test_url + '/preamble/2016_02749/intro')
        html = self.driver.find_element_by_tag_name('html')
        self.assertIn('Addition of a Subsurface Intrusion Component',
                      html.text)



"""Macros used to make a few common tasks in templates easier"""
from django import template

register = template.Library()


@register.inclusion_tag('regulations/macros/external_link.html')
def external_link(url, text, classes="", title=""):
    return {"url": url, "text": text, "classes": classes, "title": title}


@register.inclusion_tag('regulations/macros/search_for.html')
def search_for(terms, reg, version, text=None):
    if text is None:
        text = terms
    return {"terms": terms, "reg": reg, "version": version, "text": text}



# vim: set fileencoding=utf-8
from unittest import TestCase
from regulations.generator import title_parsing


class RegTest(TestCase):
    def test_try_split(self):
        self.assertEqual(
            ['a', 'xb'],
            title_parsing.try_split('a:xb', ('|', ':', 'x')))

    def test_appendix_supplement_ap(self):
        elements = title_parsing.appendix_supplement({
            'index': ['204', 'A2'],
            'title': 'Appendix A2 to 204-First Appendix'})
        self.assertTrue(elements['is_appendix'])
        elements = title_parsing.appendix_supplement({
            'index': ['204', 'A'],
            'title': 'Appendix A to 204-First Appendix'})
        self.assertTrue(elements['is_appendix'])

    def test_section(self):
        elements = title_parsing.section({
            'index': ['204', '4'],
            'title': u'§ 204.4 Sauce'})

        self.assertTrue(elements['is_section'])
        self.assertEquals('Sauce', elements['sub_label'])

        elements = title_parsing.section({
            'index': ['204', '4'],
            'title': u'§ 204.4 [Reserved]'})

        self.assertTrue(elements['is_section'])
        self.assertEquals('[Reserved]', elements['sub_label'])



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from unittest import TestCase

from nose.tools import assert_equal, assert_is_none
from mock import patch

from fr_notices import navigation


class PreambleToCTests(TestCase):
    def setUp(self):
        self.nodes = [{
            'title': 'one',
            'label': ['doc_id', 'I'],
            'children': [{
                'title': 'two',
                'label': ['doc_id', 'I', 'A'],
            }]
        }, {
            'title': 'three',
            'label': ['doc_id', 'II'],
        }]

    def test_preamble_toc(self):
        toc = navigation.make_preamble_nav(self.nodes)
        self.assertEqual(toc, [
            navigation.NavItem(
                url='/preamble/doc_id/I',
                title=navigation.Title('one', 'I', 'one'),
                markup_id='doc_id-preamble-doc_id-I',
                children=[navigation.NavItem(
                    url='/preamble/doc_id/I#doc_id-I-A',
                    title=navigation.Title('two', 'A', 'two'),
                    markup_id='doc_id-preamble-doc_id-I-A',
                )]
            ),
            navigation.NavItem(
                url='/preamble/doc_id/II',
                title=navigation.Title('three', 'II', 'three'),
                markup_id='doc_id-preamble-doc_id-II',
            )
        ])

    def test_max_depth(self):
        toc = navigation.make_preamble_nav(self.nodes, max_depth=1)
        self.assertEqual(toc[0].children, [])

    def test_footer(self):
        toc = navigation.make_preamble_nav(self.nodes)

        nav = navigation.footer(toc, [], 'doc_id-preamble-doc_id-I')
        assert_equal(nav['next'].section_id, 'doc_id-preamble-doc_id-II')
        assert_is_none(nav['previous'])

        nav = navigation.footer(toc, [], 'doc_id-preamble-doc_id-II')
        assert_equal(nav['previous'].section_id, 'doc_id-preamble-doc_id-I')
        assert_is_none(nav['next'])


class CFRChangeBuilderTests(TestCase):
    @patch('fr_notices.navigation.fetch_toc')
    @patch('fr_notices.navigation.utils.regulation_meta')
    def test_make_cfr_change_nav(self, fetch_meta, fetch_toc):
        """Add amendments for two different CFR parts. Verify that the table
        of contents contains only the changed data. Also validate that changes
        to subparts do not include a ToC entry"""
        version_info = {'111': {'left': 'v1', 'right': 'v2'},
                        '222': {'left': 'vold', 'right': 'vnew'}}
        fetch_toc.return_value = [
            dict(index=['111', '1'], title='§ 111.1 Something'),
            dict(index=['111', '1', 'a'], title='1 a'),
            dict(index=['111', '2'], title='§ 111.2 Else Here'),
            dict(index=['111', '3'], title='Unparsable'),
            # Realistically we wouldn't have these together, but it makes
            # mocking these results easier
            dict(index=['222', '4'], title='Section 4')
        ]
        fetch_meta.return_value = dict(cfr_title_number='99',
                                       statutory_name='Some name')
        toc = navigation.make_cfr_change_nav('docdoc', version_info, [
            dict(cfr_part='111', instruction='1. inst1', authority='auth1'),
            # subpart change -- doesn't affect ToC
            dict(cfr_part='111', instruction='2. inst2',
                 changes=[['111-Subpart-A', []]]),
            dict(cfr_part='111', instruction='2. inst2',
                 # The second element of each pair would be
                 # non-empty in realistic scenarios
                 changes=[['111-1', []], ['111-3-b', []]]),
            # only authority change
            dict(cfr_part='222', instruction='3. inst3', authority='auth2')
        ])

        self.assertEqual(toc, [
            navigation.NavItem(
                url='/preamble/docdoc/cfr_changes/111',
                title=navigation.Title('Authority', '99 CFR 111', 'Authority'),
                markup_id='docdoc-cfr-111',
                category='99 CFR 111',
                section_id=''),
            navigation.NavItem(
                url='/preamble/docdoc/cfr_changes/111-1',
                title=navigation.Title(
                    '§ 111.1 Something', '§ 111.1', 'Something'),
                markup_id='docdoc-cfr-111-1',
                category='99 CFR 111'),
            navigation.NavItem(
                url='/preamble/docdoc/cfr_changes/111-3',
                title=navigation.Title('Unparsable', 'Unparsable'),
                markup_id='docdoc-cfr-111-3',
                category='99 CFR 111'),
            navigation.NavItem(
                url='/preamble/docdoc/cfr_changes/222',
                title=navigation.Title('Authority', '99 CFR 222', 'Authority'),
                markup_id='docdoc-cfr-222',
                category='99 CFR 222',
                section_id='')
        ])

    def test_footer(self):
        toc = [
            navigation.NavItem(
                url='/preamble/2016_02749/cfr_changes/478',
                title=navigation.Title('Authority', '27 CFR 478', 'Authority'),
                markup_id='2016_02749-cfr-478',
                category='27 CFR 478'),
            navigation.NavItem(
                url='/preamble/2016_02749/cfr_changes/478-99',
                title=navigation.Title(
                    '§ 478.99 Certain', '§ 478.99', 'Certain'),
                markup_id='2016_02749-cfr-478-99',
                category='27 CFR 478'),
            navigation.NavItem(
                url='/preamble/2016_02749/cfr_changes/478-120',
                title=navigation.Title(
                    '§ 478.120 Firearms', '§ 478.120', 'Firearms'),
                markup_id='2016_02749-cfr-478-120',
                category='27 CFR 478')
        ]

        nav = navigation.footer([], toc, '2016_02749-cfr-478')
        assert_is_none(nav['previous'])
        assert_equal(nav['next'].section_id, '2016_02749-cfr-478-99')

        nav = navigation.footer([], toc, '2016_02749-cfr-478-99')
        assert_equal(nav['previous'].section_id, '2016_02749-cfr-478')
        assert_equal(nav['next'].section_id, '2016_02749-cfr-478-120')

        nav = navigation.footer([], toc, '2016_02749-cfr-478-120')
        assert_equal(nav['previous'].section_id, '2016_02749-cfr-478-99')
        assert_is_none(nav['next'])



"""The table of contents needed isn't quite the same as the toc provided by
the api. We need to modify it a bit to group subparts, subterps, etc. These
modifications, then, are used for navigation, citations, and the TOC
layer"""
import logging

from regulations.generator import api_reader, title_parsing


logger = logging.getLogger(__name__)


def fetch_toc(reg_part, version, flatten=False):
    """Fetch the toc, transform it into a list usable by navigation, etc."""
    toc = api_reader.ApiReader().layer('toc', 'cfr', reg_part, version)
    if toc is None:
        logger.warning("404 when fetching TOC for %s@%s", reg_part, version)
        toc = []

    toc_list = []
    if reg_part in toc:
        for data in toc[reg_part]:
            if 'Subpart' in data['index']:
                toc_list.append(toc_subpart(data, toc_list, toc))
            elif 'Subjgrp' in data['index']:
                toc_list.append(toc_subjgrp(data, toc_list, toc))
            elif 'Interp' in data['index']:
                toc_list.append(toc_interp(data, toc_list, toc))
            else:
                toc_list.append(toc_sect_appendix(data, toc_list))
    if flatten:
        flattened = []
        for el in toc_list:
            if 'sub_toc' in el:
                flattened.extend(el['sub_toc'])
            else:
                flattened.append(el)
        return flattened
    return toc_list


def toc_sect_appendix(data, so_far):
    """Transforms normal sections and appendices"""
    title_data = title_parsing.section(data)
    if title_data:
        data.update(title_data)

    title_data = title_parsing.appendix_supplement(data)
    if title_data:
        data.update(title_data)

    if 'label' not in data:
        data['label'] = data['title']

    if data.get('is_appendix'):
        seen_appendix = any(el.get('is_appendix') for el in so_far)
        data['is_first_appendix'] = not seen_appendix
    return data


def toc_subpart(data, so_far, toc):
    """Transforms a subpart, giving it sectional children"""
    element = {
        'label': ' '.join(data['index'][1:]),
        'sub_label': data['title'],
        'index': data['index'],
        'section_id': '-'.join(data['index']),
        'is_subpart': True,
        'sub_toc': []
    }
    for sub in toc.get('-'.join(data['index']), []):
        element['sub_toc'].append(toc_sect_appendix(sub, so_far))
    return element


def toc_subjgrp(data, so_far, toc):
    """Transforms a subpart, giving it sectional children"""
    element = {
        'label': data["title"],
        'sub_label': "",
        'index': data['index'],
        'section_id': '-'.join(data['index']),
        'is_subjgrp': True,
        'sub_toc': []
    }
    for sub in toc.get('-'.join(data['index']), []):
        element['sub_toc'].append(toc_sect_appendix(sub, so_far))
    return element


def toc_interp(data, so_far, toc):
    """Transforms a subpart, expanding it into subterps (collections of
    interpreted subparts, empty part, and appendices"""
    segments = title_parsing.try_split(data['title'])
    if not segments:
        segments = 'Supplement I', ''
    element = {
        'label': segments[0],
        'sub_label': segments[1],
        'index': data['index'],
        'section_id': '-'.join(data['index']),
        'is_supplement': True,
        'sub_toc': []
    }

    reg_part = data['index'][0]
    element['sub_toc'].extend(intro_interps(toc, reg_part))
    element['sub_toc'].extend(subterps(so_far, reg_part))
    return element


def intro_interps(toc, reg_part):
    """Logic to fill in any introduction headers for the entire
    interpretations. Note that at some point, we might have headers randomly
    appear elsewhere in the interpretations, unrelated to a specific
    section. That's a @todo."""
    elements = []
    for el in toc.get(reg_part + '-Interp', []):
        if el['index'][1] == 'Interp':
            elements.append({
                'label': 'Interpretations',
                'sub_label': el['title'],
                'index': el['index'],
                'section_id': '-'.join(el['index'])})
    return elements


def subterps(so_far, reg_part):
    """Logic to build subterps, collections of interpretations for subparts,
    the empty subpart, or appendices"""
    elements = []
    found_subpart = False
    found_appendix = False
    for el in so_far:
        if el.get('is_subpart'):
            found_subpart = True
            index = el['index'] + ['Interp']
            elements.append({
                'label': el['label'],
                'sub_label': el['sub_label'],
                'index': index,
                'is_subterp': True,
                'section_id': '-'.join(index)
            })
        elif el.get('is_appendix') and not found_appendix:
            found_appendix = True
            index = el['index'][:1] + ['Appendices', 'Interp']
            elements.append({
                'label': 'Appendices',
                'index': index,
                'is_subterp': True,
                'section_id': '-'.join(index)
            })

    if not found_subpart:   # Add the empty subpart
        index = [reg_part, 'Subpart', 'Interp']
        elements.insert(0, {
            'label': 'Regulation Text',
            'index': index,
            'is_subterp': True,
            'section_id': '-'.join(index)
        })
    return elements



# vim: set encoding=utf-8
from collections import defaultdict

from django.template import Context, loader
from django.views.generic.base import TemplateView

from regulations.generator import api_reader, generator, notices
from regulations.generator.layers.utils import convert_to_python
from regulations.generator.node_types import label_to_text
from regulations.generator.section_url import SectionUrl
from regulations.views import error_handling


class ParagraphSXSView(TemplateView):
    """ Given a regulation paragraph and a Federal Register notice number,
    display the appropriate section by section analyses."""
    def __init__(self):
        self.footnote_tpl = loader.get_template(
            "regulations/layers/sxs-footnotes.html")

    def get_template_names(self):
        """ The disclaimer that exists on this page can be over ridden. If an
        agency specfic disclaimer is provided, use that. """

        return ['regulations/sxs_with_disclaimer.html',
                'regulations/paragraph-sxs.html']

    def get(self, request, *args, **kwargs):
        """Override this method so that we can grab the GET variables"""
        kwargs['version'] = request.GET.get('from_version')
        kwargs['back_url'] = SectionUrl.of(
            kwargs['label_id'].split('-'), kwargs['version'], sectional=True)
        kwargs['fr_page'] = request.GET.get('fr_page')
        if kwargs['fr_page'] and kwargs['fr_page'].isdigit():
            kwargs['fr_page'] = int(kwargs['fr_page'])
        return super(ParagraphSXSView, self).get(request, *args,
                                                 **kwargs)

    def further_analyses(self, label_id, notice_id, fr_page, version):
        """Grab other analyses for this same paragraph (limiting to those
           visible from this regulation version.) Make them in descending
           order"""
        sxs_layer_data = api_reader.ApiReader().layer(
            'analyses', 'cfr', label_id, version)

        if label_id not in sxs_layer_data:
            return []
        else:
            return [convert_to_python(a)
                    for a in reversed(sxs_layer_data[label_id])
                    if (a['reference'] != [notice_id, label_id]
                        or a['fr_page'] != fr_page)]

    def footnote_refs(self, sxs):
        """Add footnote references to paragraph text"""
        refs = sxs.get('footnote_refs', [])
        ref_dict = defaultdict(list)
        for ref in refs:
            ref_dict[ref['paragraph']].append(ref)

        for p_idx in range(len(sxs['paragraphs'])):
            shift = 0
            for ref in ref_dict[p_idx]:
                p = sxs['paragraphs'][p_idx]
                rendered = self.footnote_tpl.render(Context({'footnote': ref}))
                offset = ref['offset'] + shift
                sxs['paragraphs'][p_idx] = p[:offset] + rendered
                sxs['paragraphs'][p_idx] += p[offset:]
                shift += len(rendered)
        for child in sxs['children']:
            self.footnote_refs(child)

    def footnotes(self, notice, sxs):
        """Data for footnotes (which are referenced in the paragraph text)"""
        feet = []
        for ref in sxs.get('footnote_refs', []):
            ref['text'] = notice['footnotes'][ref['reference']]
            feet.append(ref)
        for child in sxs['children']:
            feet = feet + self.footnotes(notice, child)
        return feet

    def get_context_data(self, **kwargs):
        context = super(ParagraphSXSView, self).get_context_data(**kwargs)

        label_id = context['label_id']
        notice_id = context['notice_id']
        fr_page = context.get('fr_page')

        notice = generator.get_notice(notice_id)
        if not notice:
            raise error_handling.MissingContentException()
        notice = convert_to_python(notice)

        paragraph_sxs = generator.get_sxs(label_id, notice, fr_page)

        if paragraph_sxs is None:
            raise error_handling.MissingContentException()

        notices.add_depths(paragraph_sxs, 3)

        paragraph_sxs['children'] =\
            notices.filter_labeled_children(paragraph_sxs)
        self.footnote_refs(paragraph_sxs)

        context['sxs'] = paragraph_sxs
        # Template assumes a single label
        context['sxs']['label'] = context['label_id']
        context['sxs']['header'] = label_to_text(label_id.split('-'),
                                                 include_marker=True)
        context['sxs']['all_footnotes'] = self.footnotes(notice, paragraph_sxs)
        context['notice'] = notice
        context['further_analyses'] = self.further_analyses(
            label_id, notice_id, paragraph_sxs['page'], context['version'])

        return context



from django.template import loader, Context

from regulations.generator.layers.base import LayerBase


# Does not extend InlineLayer as this retrieves its data in a different way
class DefinedLayer(LayerBase):
    shorthand = 'defined'
    data_source = 'terms'
    layer_type = LayerBase.INLINE

    def __init__(self, layer):
        self.layer = layer
        self.template = loader.get_template('regulations/layers/defining.html')

    def apply_layer(self, text, text_index):
        """Catch all terms which are defined in this paragraph, replace them
        with a span"""
        layer_pairs = []
        for ref_struct in self.layer['referenced'].values():
            if text_index == ref_struct['reference']:
                pos = tuple(ref_struct['position'])
                original = text[pos[0]:pos[1]]
                context = Context({'term': original})
                replacement = self.template.render(context)
                replacement = replacement.strip('\n')
                layer_pairs.append((original, replacement, pos))
        return layer_pairs



from six.moves.urllib_parse import urlencode

from django.conf import settings
from django.core.urlresolvers import reverse


def eregs_globals(request):
    env = 'source' if getattr(settings, 'JS_DEBUG', False) else 'built'
    prefix = reverse('regulation_landing_view', kwargs={'label_id': '9999'})
    prefix = prefix.replace('9999', '')
    analytics = getattr(settings, 'ANALYTICS', {})
    if 'DAP' in analytics:
        analytics['DAP']['DAP_URL_PARAMS'] = create_dap_url_params(
            analytics['DAP'])
    return {
        'EREGS_GLOBALS': {
            'ENV': env,
            'APP_PREFIX': prefix,
            'ANALYTICS': analytics,
        },
    }


def create_dap_url_params(dap_settings):
    """ Create the DAP url string to append to script tag """
    dap_params = {}
    if 'AGENCY' in dap_settings and dap_settings['AGENCY']:
        dap_params['agency'] = dap_settings['AGENCY']
        if 'SUBAGENCY' in dap_settings and dap_settings['SUBAGENCY']:
            dap_params['subagency'] = dap_settings['SUBAGENCY']

    return urlencode(dap_params)



# vim: set fileencoding=utf-8
from unittest import TestCase

from mock import patch

from regulations.generator import toc


class TocTest(TestCase):
    def test_toc_sect_appendix(self):
        data = {'title': u'§ 1026.1 - Authority', 'index': ['1026', '1']}
        result = toc.toc_sect_appendix(data, [])
        self.assertEqual(u'1026.1', result['label'])
        self.assertEqual('Authority', result['sub_label'])
        self.assertTrue(result['is_section'])

        data = {'title': u'§ 1026.1 [Reserved]', 'index': ['1026', '1']}
        result = toc.toc_sect_appendix(data, [])
        self.assertEqual(u'1026.1', result['label'])
        self.assertEqual('[Reserved]', result['sub_label'])
        self.assertTrue(result['is_section'])

        data = {'title': u'Appendix A', 'index': ['1026', 'A']}
        result = toc.toc_sect_appendix(data, [])
        self.assertEqual('Appendix A', result['label'])
        self.assertTrue(result['is_first_appendix'])
        self.assertTrue(result['is_appendix'])

        data = {'title': u'Appendix B [Reserved]', 'index': ['1026', 'A']}
        result = toc.toc_sect_appendix(data, [])
        self.assertEqual('Appendix B', result['label'])
        self.assertEqual('[Reserved]', result['sub_label'])
        self.assertTrue(result['is_first_appendix'])
        self.assertTrue(result['is_appendix'])

        data = {'title': u'Appendix B [Reserved]', 'index': ['1026', 'A']}
        result = toc.toc_sect_appendix(data, [{'is_appendix': True}])
        self.assertEqual('Appendix B', result['label'])
        self.assertEqual('[Reserved]', result['sub_label'])
        self.assertFalse(result['is_first_appendix'])
        self.assertTrue(result['is_appendix'])

    def test_toc_subpart(self):
        layer = {'1001-Subpart-A': [{'title': u'§ 1001.1 - Content',
                                     'index': ['1001', '1']},
                                    {'title': u'§ 1001.2 - Other',
                                     'index': ['1001', '2']}]}
        data = {'title': u'General', 'index': ['1001', 'Subpart', 'A']}
        result = toc.toc_subpart(data, [], layer)
        self.assertEqual('Subpart A', result['label'])
        self.assertEqual('General', result['sub_label'])
        self.assertEqual(['1001', 'Subpart', 'A'], result['index'])
        self.assertEqual('1001-Subpart-A', result['section_id'])
        self.assertTrue(result.get('is_subpart'))
        self.assertEqual(2, len(result['sub_toc']))

        s1, s2 = result['sub_toc']
        self.assertEqual('Content', s1['sub_label'])
        self.assertEqual(['1001', '1'], s1['index'])
        self.assertEqual('Other', s2['sub_label'])
        self.assertEqual(['1001', '2'], s2['index'])

    def test_toc_subpart_with_nondigits(self):
        layer = {'1001-Subpart-A': [{'title': u'§ 1001.1a - Content',
                                     'index': ['1001', '1a']},
                                    {'title': u'§ 1001.2 - Other',
                                     'index': ['1001', '2']}]}
        data = {'title': u'General', 'index': ['1001', 'Subpart', 'A']}
        result = toc.toc_subpart(data, [], layer)
        self.assertEqual('Subpart A', result['label'])
        self.assertEqual('General', result['sub_label'])
        self.assertEqual(['1001', 'Subpart', 'A'], result['index'])
        self.assertEqual('1001-Subpart-A', result['section_id'])
        self.assertTrue(result.get('is_subpart'))
        self.assertEqual(2, len(result['sub_toc']))

        s1, s2 = result['sub_toc']
        self.assertEqual('Content', s1['sub_label'])
        self.assertEqual(['1001', '1a'], s1['index'])
        self.assertEqual('Other', s2['sub_label'])
        self.assertEqual(['1001', '2'], s2['index'])

    def test_toc_subpart_reserved(self):
        layer = {'1001-Subpart-A': [{'title': '1001.1 - Content',
                                     'index': ['1001', '1']},
                                    {'title': '1001.2 - Other',
                                     'index': ['1001', '2']}]}
        data = {'title': u'[Reserved]', 'index': ['1001', 'Subpart', 'B']}
        result = toc.toc_subpart(data, [], layer)
        self.assertEqual('Subpart B', result['label'])
        self.assertEqual('[Reserved]', result['sub_label'])
        self.assertEqual(['1001', 'Subpart', 'B'], result['index'])
        self.assertEqual('1001-Subpart-B', result['section_id'])
        self.assertTrue(result.get('is_subpart'))
        self.assertEqual(0, len(result['sub_toc']))

    def test_toc_interp(self):
        so_far = [
            {'index': ['1001', '1']},
            {'index': ['1001', '2']},
            {'index': ['1001', 'A'], 'is_appendix': True},
            {'index': ['1001', 'B'], 'is_appendix': True}]
        data = {'title': 'Supplement II - Official Interps',
                'index': ['1001', 'Interp']}
        toc_data = {'1001-Interp': []}
        result = toc.toc_interp(data, so_far, toc_data)
        self.assertEqual('Supplement II', result['label'])
        self.assertEqual('Official Interps', result['sub_label'])
        self.assertTrue(result.get('is_supplement'))
        self.assertEqual(2, len(result['sub_toc']))

        reg, app = result['sub_toc']
        self.assertEqual('Regulation Text', reg['label'])
        self.assertTrue(reg.get('is_subterp'))
        self.assertEqual(['1001', 'Subpart', 'Interp'], reg['index'])
        self.assertEqual('1001-Subpart-Interp', reg['section_id'])

        self.assertEqual('Appendices', app['label'])
        self.assertTrue(app.get('is_subterp'))
        self.assertEqual(['1001', 'Appendices', 'Interp'], app['index'])
        self.assertEqual('1001-Appendices-Interp', app['section_id'])

        so_far = [
            {'index': ['1001', 'Subpart', 'C'], 'is_subpart': True,
             'label': 'Subpart C', 'sub_label': 'Awesome Sauce',
             'sub_toc': [{'index': ['1001', '1']}, {'index': ['1001', '2']}]}]
        data = {'title': 'Unparsable',
                'index': ['1001', 'Interp']}
        result = toc.toc_interp(data, so_far, toc_data)
        self.assertEqual('Supplement I', result['label'])
        self.assertEqual('', result['sub_label'])
        self.assertTrue(result.get('is_supplement'))
        self.assertEqual(1, len(result['sub_toc']))

        subpart = result['sub_toc'][0]
        self.assertEqual('Subpart C', subpart['label'])
        self.assertEqual('Awesome Sauce', subpart['sub_label'])
        self.assertTrue(subpart.get('is_subterp'))
        self.assertEqual(['1001', 'Subpart', 'C', 'Interp'], subpart['index'])
        self.assertEqual('1001-Subpart-C-Interp', subpart['section_id'])

        toc_data = {'1001-Interp': [{'title': 'Intro',
                                     'index': ['1001', 'Interp', 'h1']},
                                    {'title': 'Section 1',
                                     'index': ['1001', '1', 'Interp']}]}
        result = toc.toc_interp(data, so_far, toc_data)
        self.assertEqual(2, len(result['sub_toc']))

        h1, subpart = result['sub_toc']
        self.assertEqual('Interpretations', h1['label'])
        self.assertEqual('Intro', h1['sub_label'])
        self.assertFalse(h1.get('is_subterp', False))
        self.assertEqual('1001-Interp-h1', h1['section_id'])
        self.assertEqual(['1001', 'Interp', 'h1'], h1['index'])
        self.assertTrue(subpart.get('is_subterp'))
        self.assertEqual(['1001', 'Subpart', 'C', 'Interp'], subpart['index'])
        self.assertEqual('1001-Subpart-C-Interp', subpart['section_id'])

    @patch('regulations.generator.toc.api_reader')
    def test_fetch_toc_404(self, api_reader):
        """Should not crash if there is no TOC data"""
        api_reader.ApiReader.return_value.layer.return_value = None
        self.assertEqual([], toc.fetch_toc('111', 'vvv'))
        self.assertEqual([], toc.fetch_toc('111', 'vvv', True))



from unittest import TestCase

from mock import patch
from django.http import Http404
from django.test import RequestFactory

from regulations.generator.layers.layers_applier import (
    InlineLayersApplier, ParagraphLayersApplier, SearchReplaceLayersApplier)
from regulations.generator.node_types import INTERP
from regulations.views import partial_interp


class PartialInterpViewTest(TestCase):
    @patch('regulations.views.partial.generator')
    def test_get_context_data(self, generator):
        generator.get_tree_paragraph.return_value = {
            'text': 'Some Text',
            'children': [],
            'label': ['867', '53', 'q', 'Interp'],
            'node_type': INTERP
        }
        request = RequestFactory().get('/fake-path')
        view = partial_interp.PartialInterpView.as_view(
            appliers=(InlineLayersApplier(), ParagraphLayersApplier(),
                      SearchReplaceLayersApplier()))
        response = view(request, label_id='lablab', version='verver')
        self.assertEqual(response.context_data['c']['node_type'], INTERP)
        self.assertEqual(response.context_data['c']['children'],
                         [generator.get_tree_paragraph.return_value])
        self.assertFalse(response.context_data['inline'])

        view = partial_interp.PartialInterpView.as_view(
            inline=True,
            appliers=(InlineLayersApplier(), ParagraphLayersApplier(),
                      SearchReplaceLayersApplier()))
        response = view(request, label_id='lablab', version='verver')
        self.assertEqual(response.context_data['c']['node_type'], INTERP)
        self.assertEqual(response.context_data['c']['children'],
                         [generator.get_tree_paragraph.return_value])
        self.assertTrue(response.context_data['inline'])


class PartialSubterpViewTest(TestCase):
    @patch('regulations.views.partial.generator')
    @patch('regulations.views.partial_interp.generator')
    @patch('regulations.views.partial_interp.filter_by_subterp')
    @patch('regulations.views.partial_interp.generate_html')
    @patch('regulations.views.partial.navigation')
    def test_get_context_data(self, nav, genhtml, filter_by_subterp,
                              interp_generator, partial_generator):
        get_appliers = partial_generator.LayerCreator.return_value.get_appliers
        get_appliers.return_value = (None, None, None)
        interp_generator.get_tree_paragraph.return_value = None
        nav.nav_sections.return_value = None, None

        request = RequestFactory().get('/fake-path')
        view = partial_interp.PartialSubterpView.as_view()
        try:
            view(request, label_id='lablab', version='verver')
            self.assertTrue(False)
        except Http404:
            pass

        interp_generator.get_tree_paragraph.return_value = {'children': []}
        filter_by_subterp.return_value = []
        view = partial_interp.PartialSubterpView.as_view()
        try:
            view(request, label_id='lablab', version='verver')
            self.assertTrue(False)
        except Http404:
            pass

        filter_by_subterp.return_value = ['sec1', 'sec2', 'sec3']
        view(request, label_id='lablab', version='verver')
        self.assertTrue(genhtml.called)
        args = genhtml.call_args[0]
        self.assertEqual({'children': ['sec1', 'sec2', 'sec3'],
                         'label': ['lablab']}, args[0])



import abc


class LayerBase(object):
    """Base class for most layers; each layer contains information which is
    added on top of the regulation, such as definitions, internal citations,
    keyterms, etc."""
    __metaclass__ = abc.ABCMeta

    # @see layer_type
    INLINE = 'inline'
    PARAGRAPH = 'paragraph'
    SEARCH_REPLACE = 'search_replace'

    @abc.abstractproperty
    def shorthand(self):
        """A short description for this layer. This is used in query strings
        and the like to define which layers should be used"""
        raise NotImplementedError

    @abc.abstractproperty
    def data_source(self):
        """Data is pulled from the API; this field indicates the name of the
        endpoint to pull data from"""
        raise NotImplementedError

    @abc.abstractproperty
    def layer_type(self):
        """Layer data can be applied in a few ways, attaching itself to a
        node, replacing text based on offset, or replacing text based on
        searching. Which type is this layer?"""
        raise NotImplementedError


class InlineLayer(LayerBase):
    """Represents a layer which replaces text by looking at offsets"""
    layer_type = LayerBase.INLINE

    @abc.abstractmethod
    def replacement_for(self, original, data):
        """Given the original text and the relevant data from a layer, create
        a (string) replacement, by, for example, running the data through a
        template"""
        raise NotImplementedError

    def apply_layer(self, text, label_id):
        """Entry point when processing the regulation tree. Given the node's
        text and its label_id, yield all replacement text"""
        data_with_offsets = ((entry, start, end)
                             for entry in self.layer.get(label_id, [])
                             for (start, end) in entry['offsets'])
        for data, start, end in data_with_offsets:
            start, end = int(start), int(end)
            original = text[start:end]
            replacement = self.replacement_for(original, data)
            yield (original, replacement, (start, end))


class SearchReplaceLayer(LayerBase):
    """Represents a layer which replaces text by searching for and replacing a
    specific substring. Also accounts for the string appearing multiple times
    (via the 'locations' field)"""
    layer_type = LayerBase.SEARCH_REPLACE
    _text_field = 'text'    # All but key terms follow this convention...

    @abc.abstractmethod
    def replacements_for(self, text, data):
        """Given the original text and the relevant data from a layer, create
        a (string) replacement, by, for example, running the data through a
        template. Returns a generator"""
        raise NotImplementedError

    def apply_layer(self, label_id):
        """Entry point when processing the regulation tree. Given the node's
        label_id, attempt to find relevant layer data in self.layer"""
        for entry in self.layer.get(label_id, []):
            text = entry[self._text_field]
            for replacement in self.replacements_for(text, entry):
                yield (text, replacement, entry['locations'])



# -*- coding: utf-8 -*-

from mock import patch
from unittest import TestCase
from datetime import date, timedelta

from nose.tools import assert_equal
from django.http import Http404
from django.test import RequestFactory, override_settings

from fr_notices.navigation import make_preamble_nav
from regulations.generator.layers import diff_applier, layers_applier
from regulations.views import preamble
from regulations.views.preamble import CommentState


class PreambleViewTests(TestCase):
    _mock_preamble = dict(text='1', label=['1'], node_type='', children=[
        dict(text='2', label=['1', 'c'], node_type='', children=[
            dict(text='3', label=['1', 'c', 'i'], node_type='', children=[]),
            dict(text='4', label=['1', 'c', 'x'], node_type='', children=[])
        ]),
        dict(text='5', label=['1', '1'], node_type='', children=[])
    ])

    def test_find_subtree(self):
        """When a node is present in a tree, we should be able to find it.
        When it is not, we should get None"""
        root = self._mock_preamble
        fn = preamble.find_subtree
        self.assertEqual(fn(root, ['1'])['text'], '1')
        self.assertEqual(fn(root, ['1', 'c'])['text'], '2')
        self.assertEqual(fn(root, ['1', 'c', 'i'])['text'], '3')
        self.assertEqual(fn(root, ['1', 'c', 'x'])['text'], '4')
        self.assertEqual(fn(root, ['1', '1'])['text'], '5')
        self.assertIsNone(fn(root, ['2']))
        self.assertIsNone(fn(root, ['1', '2']))
        self.assertIsNone(fn(root, ['1', 'c', 'r']))
        self.assertIsNone(fn(root, ['1', 'c', 'i', 'r']))

    @patch('fr_notices.navigation.CFRChangeBuilder')
    @patch('regulations.generator.generator.api_reader')
    @patch('regulations.views.preamble.ApiReader')
    def test_get_integration(self, ApiReader, api_reader, CFRChangeBuilder):
        """Verify that the contexts are built correctly before being sent to
        the template. AJAX/partial=true requests should only get the inner
        context (i.e. no UI-related context)"""
        ApiReader.return_value.preamble.return_value = self._mock_preamble
        api_reader.ApiReader.return_value.layer.return_value = {
            '1-c-x': ['something']
        }
        view = preamble.PreambleView.as_view()

        path = '/preamble/1/c/x?layers=meta'
        response = view(RequestFactory().get(path), paragraphs='1/c/x')
        self.assertEqual(
            response.context_data['sub_context']['node']['text'], '4')
        self.assertEqual(
            response.context_data['sub_context']['node']['children'], [])
        # layer data is present
        self.assertEqual(
            response.context_data['sub_context']['node']['meta'], 'something')
        self.assertEqual(
            response.context_data['preamble_toc'],
            make_preamble_nav(self._mock_preamble['children']),
        )
        self.assertNotIn('node', response.context_data)

        response = view(RequestFactory().get(path + '&partial=true'),
                        paragraphs='1/c/x')
        self.assertIn('sub_context', response.context_data)
        self.assertEqual(
            response.context_data['sub_context']['node']['text'],
            '4',
        )

        request = RequestFactory().get(
            path, HTTP_X_REQUESTED_WITH='XMLHttpRequest')
        response = view(request, paragraphs='1/c/x')
        self.assertIn('sub_context', response.context_data)
        self.assertEqual(
            response.context_data['sub_context']['node']['text'],
            '4',
        )

    @override_settings(
        PREAMBLE_INTRO={'1': {'meta': {
            'publication_date': '2001-01-01',
            'comments_close': (date.today() + timedelta(days=1)).isoformat()
        }}})
    @patch('regulations.views.preamble.ApiReader')
    def test_comments_open_from_settings(self, ApiReader):
        """
        Mock the PREAMBLE_INTRO data from settings for this test of the
        comments being open.
        """
        _, meta, _ = preamble.notice_data('1')

        assert_equal(meta['comment_state'], CommentState.OPEN)

    def _setup_mock_response(self, ApiReader, **kwargs):
        """Mock the ApiReader response, replacing meta data fields with
        kwargs"""
        ApiReader.return_value.preamble.return_value = self._mock_preamble
        notice = {
            "action": "Proposed rule",
            "agencies": ["Environmental Protection Agency"],
            "cfr_title": 40,
            "cfr_parts": ["300"],
            "comments_close": "2011-09-09",
            "dockets": ["EPA-HQ-SFUND-2010-1086",
                        "FRL-9925-69-OLEM"],
            "primary_agency": "Environmental Protection Agency",
            "title": ("Addition of a Subsurface Intrusion Component to the "
                      "Hazard Ranking System"),
            "publication_date": "2011-02-02",
            "regulatory_id_numbers": ["2050-AG67"],
        }
        notice.update(kwargs)
        ApiReader.return_value.notice.return_value = notice

    @patch('regulations.views.preamble.ApiReader')
    def test_comments_open(self, ApiReader):
        future = date.today() + timedelta(days=10)
        self._setup_mock_response(ApiReader, comments_close=future.isoformat())
        _, meta, _ = preamble.notice_data('1')
        assert_equal(meta['comment_state'], CommentState.OPEN)

    @patch('regulations.views.preamble.ApiReader')
    def test_comments_prepub(self, ApiReader):
        future = date.today() + timedelta(days=10)
        self._setup_mock_response(ApiReader,
                                  publication_date=future.isoformat())
        _, meta, _ = preamble.notice_data('1')
        assert_equal(meta['comment_state'], CommentState.PREPUB)

    @patch('regulations.views.preamble.ApiReader')
    def test_comments_closed(self, ApiReader):
        self._setup_mock_response(ApiReader)
        _, meta, _ = preamble.notice_data('1')
        assert_equal(meta['comment_state'], CommentState.CLOSED)

    @patch('fr_notices.navigation.CFRChangeBuilder')
    @patch('regulations.generator.generator.api_reader')
    @patch('regulations.views.preamble.ApiReader')
    def test_get_top_level_redirect(self, ApiReader, api_reader,
                                    CFRChangeBuilder):
        ApiReader.return_value.preamble.return_value = self._mock_preamble
        api_reader.ApiReader.return_value.layer.return_value = {
            '1-c-x': ['something']
        }
        view = preamble.PreambleView.as_view()

        path = '/preamble/1'
        response = view(RequestFactory().get(path), paragraphs='1')
        assert_equal(response.status_code, 302)
        assert_equal(response.get('Location'), '/preamble/1/c')

    @patch('regulations.views.preamble.ApiReader')
    def test_get_404(self, ApiReader):
        """When a requested doc is not present, we should return a 404"""
        ApiReader.return_value.preamble.return_value = None
        view = preamble.PreambleView.as_view()
        self.assertRaises(Http404, view,
                          RequestFactory().get('/preamble/1/c/x'),
                          paragraphs='1/c/x')

    @patch('regulations.views.preamble.ApiReader')
    def test_get_subtree_404(self, ApiReader):
        """When a requested _subtree_ is not present, we should 404"""
        ApiReader.return_value.preamble.return_value = self._mock_preamble
        view = preamble.PreambleView.as_view()
        self.assertRaises(Http404, view,
                          RequestFactory().get('/preamble/1/not/here'),
                          paragraphs='1/not/here')

    @patch('regulations.views.preamble.ApiReader')
    def test_notice_data(self, ApiReader):
        """We should try to fetch data corresponding to both the Preamble and
        the Notice"""
        ApiReader.return_value.preamble.return_value = self._mock_preamble
        ApiReader.return_value.notice.return_value = {
            'publication_date': '2002-02-02',
            'comments_close': '2003-03-03',
            'cfr_title': 21, 'cfr_parts': ['123']}

        for doc_id in ('123_456', '123-456'):
            preamble_, meta, notice = preamble.notice_data(doc_id)
            self.assertEqual(preamble_, self._mock_preamble)
            assert_equal(meta['comment_state'], CommentState.CLOSED)
            self.assertEqual(meta['cfr_refs'],
                             [{'title': 21, 'parts': ['123']}])
            self.assertEqual(ApiReader.return_value.preamble.call_args[0][0],
                             '123_456')
            self.assertEqual(ApiReader.return_value.notice.call_args[0][0],
                             '123-456')


class CFRChangesViewTests(TestCase):
    @patch('regulations.views.preamble.ApiReader')
    @patch('regulations.views.preamble.get_appliers')
    def test_new_regtext_changes(self, get_appliers, ApiReader):
        """We can add a whole new section without explosions"""
        amendments = [{'instruction': '3. Add subpart M',
                       'changes': [
                           ['111-Subpart-M', [{'node': {
                               'label': ['111', 'Subpart', 'M'],
                               'title': 'A New Subpart',
                               'child_labels': ['111-42', '111-43',
                                                '111-44', '111-45']}}]],
                           ['111-42', [{'some': 'thing'}]],
                           ['111-43', [{'some': 'thing'}]],
                           ['111-44', [{'some': 'thing'}]],
                           ['111-45', [{'some': 'thing'}]]]},
                      {'instruction': '4. Unrelated'}]
        version_info = {'111': {'left': '234-567', 'right': '8675-309'}}

        # Section did not exist before
        ApiReader.return_value.regulation.return_value = None
        diff = {'111-44': {'op': 'added', 'node': {
            'text': 'New node text', 'node_type': 'regtext',
            'label': ['111', '44']}}}
        get_appliers.return_value = (
            layers_applier.InlineLayersApplier(),
            layers_applier.ParagraphLayersApplier(),
            layers_applier.SearchReplaceLayersApplier(),
            diff_applier.DiffApplier(diff, '111-44'))

        result = preamble.CFRChangesView.regtext_changes_context(
            amendments, version_info, '111-44', '8675-309', 0)
        self.assertEqual(result['instructions'], ['3. Add subpart M'])
        self.assertEqual(result['tree']['marked_up'],
                         '<ins>New node text</ins>')
        self.assertEqual(1, len(result['subparts']))
        subpart_info = result['subparts'][0]
        self.assertEqual('M', subpart_info.letter)
        self.assertEqual('A New Subpart', subpart_info.title)
        self.assertEqual(2, subpart_info.idx)
        self.assertEqual(4, len(subpart_info.urls))
        self.assertIn('111-42', subpart_info.urls[0])
        self.assertIn('111-43', subpart_info.urls[1])
        self.assertIn('111-44', subpart_info.urls[2])
        self.assertIn('111-45', subpart_info.urls[3])



import json
import mock
import six

import botocore
from celery.exceptions import Retry, MaxRetriesExceededError
from django.conf import settings
from django.test import TestCase, override_settings
from nose.tools import assert_equal
from requests.exceptions import RequestException

from regulations.tasks import submit_comment, cache_pdf, SignedUrl
from regulations.models import FailedCommentSubmission


@mock.patch('regulations.tasks.submit_comment.retry')
@mock.patch('regulations.tasks.post_submission')
@mock.patch('regulations.tasks.html_to_pdf')
@mock.patch('regulations.tasks.cache_pdf')
@override_settings(
    ATTACHMENT_BUCKET='test-bucket',
    ATTACHMENT_ACCESS_KEY_ID='test-access-key',
    ATTACHMENT_SECRET_ACCESS_KEY='test-secret-key',
    ATTACHMENT_MAX_SIZE=42,
    REGS_GOV_API_URL='test-url',
    REGS_GOV_API_KEY='test-key',
)
class TestSubmitComment(TestCase):

    def setUp(self):
        self.file_handle = six.BytesIO(b"some-byte-content")
        self.comments = [
            {"id": "A1", "comment": "A simple comment", "docId": "1234_56",
             "files": []},
            {"id": "A5", "comment": "Another comment", "docId": "1234_56",
             "files": []}
        ]
        self.form = {}
        self.meta = SignedUrl('meta', 'https://s3.amazonaws.com/bucket/meta')

    def tearDown(self):
        FailedCommentSubmission.objects.all().delete()

    def test_submit_comment(self, cache_pdf, html_to_pdf, post_submission,
                            retry):
        cache_pdf.return_value = SignedUrl(
            'pdf', 'https://s3.amazonaws.com/bucket/pdf')
        html_to_pdf.return_value.__enter__.return_value = self.file_handle

        expected_result = {'trackingNumber': 'some-tracking-number'}
        post_submission.return_value.status_code = 201
        post_submission.return_value.json.return_value = expected_result

        result = submit_comment(self.comments, self.form, self.meta)

        self.assertEqual(
            result,
            {
                'trackingNumber': 'some-tracking-number',
                'pdfUrl': cache_pdf.return_value.url
            },
        )
        self.assertFalse(FailedCommentSubmission.objects.all(),
                         "No submissions saved")

    def test_failed_submit_raises_retry(self, cache_pdf, html_to_pdf,
                                        post_submission, retry):
        cache_pdf.return_value = SignedUrl(
            'pdf', 'https://s3.amazonaws.com/bucket/pdf')
        html_to_pdf.return_value.__enter__.return_value = self.file_handle

        post_submission.side_effect = RequestException

        retry.return_value = Retry()

        with self.assertRaises(Retry):
            submit_comment(self.comments, self.form, self.meta)
        self.assertFalse(FailedCommentSubmission.objects.all(),
                         "No submissions saved")

    def test_failed_submit_maximum_retries(self, cache_pdf, html_to_pdf,
                                           post_submission, retry):
        cache_pdf.return_value = SignedUrl(
            'pdf', 'https://s3.amazonaws.com/bucket/pdf')
        html_to_pdf.return_value.__enter__.return_value = self.file_handle

        post_submission.side_effect = RequestException

        retry.return_value = MaxRetriesExceededError()

        submit_comment(self.comments, self.form, self.meta)
        saved_submission = FailedCommentSubmission.objects.all()[0]
        self.assertEqual(
            json.dumps({
                'comments': self.comments,
                'form_data': self.form,
            }),
            saved_submission.body,
        )

    def test_error_fetching_from_s3_saves(self, cached_pdf, html_to_pdf,
                                          post_submission, retry):
        self.comments[1]['files'].append(
            {"name": "file_name.png", "key": "someKey"})
        retry.return_value = MaxRetriesExceededError()
        with mock.patch('regulations.tasks.boto3') as boto3:
            client = boto3.Session.return_value.client.return_value
            client.download_file.side_effect = botocore.exceptions.ClientError
            submit_comment(self.comments, self.form, self.meta)
        saved_submission = FailedCommentSubmission.objects.all()[0]
        self.assertEqual(
            json.dumps({
                'comments': self.comments,
                'form_data': self.form,
            }),
            saved_submission.body,
        )


class TestHelpers(TestCase):

    @mock.patch('regulations.tasks.SignedUrl.generate')
    @mock.patch('regulations.tasks.s3_client')
    def test_cache_pdf(self, s3_client, url_generate):
        meta = SignedUrl('meta', 'https://s3.amazonaws.com/bucket/meta')
        url_generate.return_value = SignedUrl(
            'pdf', 'https://s3.amazonaws.com/bucket/pdf')
        url = cache_pdf('content', "1234_56", meta)
        assert_equal(url, url_generate.return_value)
        s3_client.put_object.assert_any_call(
            Body=json.dumps({'pdfUrl': meta.url}),
            Bucket=settings.ATTACHMENT_BUCKET,
            Key=meta.key,
        )
        s3_client.put_object.assert_any_call(
            Body='content',
            Bucket=settings.ATTACHMENT_BUCKET,
            ContentType='application/pdf',
            ContentDisposition='attachment; filename="comment_1234_56.pdf"',
            Key=url.key,
        )



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

from operator import itemgetter

import logging

from django.http import Http404
from django.template.response import TemplateResponse
from django.views.generic.base import View

from regulations.generator.api_reader import ApiReader
from regulations.views.preamble import (
    notice_data, CommentState)


logger = logging.getLogger(__name__)


class NoticeHomeView(View):
    """
    Basic view that provides a list of regulations and notices to the context.
    """
    template_name = None  # We should probably have a default notice template.

    def get(self, request, *args, **kwargs):
        notices = ApiReader().notices().get("results", [])
        context = {}
        notices_meta = []

        for notice in notices:
            try:
                if notice.get("document_number"):
                    _, meta, _ = notice_data(notice["document_number"])
                    notices_meta.append(meta)
            except Http404:
                pass

        notices_meta = sorted(notices_meta, key=itemgetter("publication_date"),
                              reverse=True)

        context["notices"] = notices_meta
        # Django templates won't show contents of CommentState as an Enum, so:
        context["comment_state"] = {state.name: state.value for state in
                                    CommentState}

        template = self.template_name
        return TemplateResponse(request=request, template=template,
                                context=context)



from unittest import TestCase

from django.template import Template
from mock import patch

from regulations.views.partial_sxs import ParagraphSXSView
from regulations.generator.layers.utils import convert_to_python


class ParagrasphSXSViewTests(TestCase):
    @patch('regulations.views.partial_sxs.api_reader')
    def test_further_analyses(self, api_reader):
        doc1 = {'publication_date': '2009-04-05', 'fr_volume': 21,
                'fr_page': 98989, 'reference': ['doc1', '1212-31']}
        doc2 = {'publication_date': '2010-03-03', 'fr_volume': 22,
                'fr_page': 87655, 'reference': ['doc2', '1212-31']}
        doc3 = {'publication_date': '2010-10-12', 'fr_volume': 22,
                'fr_page': 90123, 'reference': ['doc3', '1212-31']}
        doc4 = {'publication_date': '2009-03-07', 'fr_volume': 21,
                'fr_page': 98888, 'reference': ['doc4', '1212-31-b']}
        api_reader.ApiReader.return_value.layer.return_value = {
            '1212-31': [doc1, doc2, doc3],
            '1212-31-b': [doc4]
        }

        psv = ParagraphSXSView()
        self.assertEqual(
            psv.further_analyses('1212-31', 'doc1', 98989, 'v1'),
            convert_to_python([doc3, doc2]))
        self.assertEqual(
            psv.further_analyses('1212-31', 'doc5', 0, 'v1'),
            convert_to_python([doc3, doc2, doc1]))
        self.assertEqual(
            psv.further_analyses('1212-31', 'doc3', 90123, 'v1'),
            convert_to_python([doc2, doc1]))

        self.assertEqual(
            psv.further_analyses('1212-31-b', 'doc3', 90123, 'v1'),
            convert_to_python([doc4]))
        self.assertEqual(psv.further_analyses('1212-31-b', 'doc4', 98888,
                                              'v1'), [])
        self.assertEqual(psv.further_analyses('1212-31-c', 'doc1', 98989,
                                              'v1'), [])

        # Same notice + label. Different page
        doc5 = {'publication_date': '2009-04-05', 'fr_volume': 21,
                'fr_page': 10101, 'reference': ['doc1', '1212-31']}
        api_reader.ApiReader.return_value.layer.return_value = {
            '1212-31': [doc1, doc2, doc3, doc5],
            '1212-31-b': [doc4]
        }
        self.assertEqual(
            psv.further_analyses('1212-31', 'doc1', 98989, 'v1'),
            convert_to_python([doc5, doc3, doc2]))
        self.assertEqual(
            psv.further_analyses('1212-31', 'doc1', 10101, 'v1'),
            convert_to_python([doc3, doc2, doc1]))
        self.assertEqual(
            psv.further_analyses('1212-31', 'doc3', 90123, 'v1'),
            convert_to_python([doc5, doc2, doc1]))

    def test_footnotes(self):
        psv = ParagraphSXSView()
        notice = {'footnotes': {
            '12': 'Twelve',
            '22': 'Twenty-two',
            '31': 'Thirty-one',
            '45': 'Forty-five'
        }}
        sxs = {
            'footnote_refs': [
                {'reference': '12', 'paragraph': 4, 'offset': 233},
                {'reference': '22', 'paragraph': 2, 'offset': 111}],
            'children': [
                {'footnote_refs': [
                    {'reference': '45', 'paragraph': 3, 'offset': 22}],
                 'children': []}]}
        footnotes = psv.footnotes(notice, sxs)
        self.assertEqual(len(footnotes), 3)
        self.assertEqual(footnotes[0]['reference'], '12')
        self.assertEqual(footnotes[0]['text'], 'Twelve')
        self.assertEqual(footnotes[1]['reference'], '22')
        self.assertEqual(footnotes[1]['text'], 'Twenty-two')
        self.assertEqual(footnotes[2]['reference'], '45')
        self.assertEqual(footnotes[2]['text'], 'Forty-five')

    def test_footnote_refs(self):
        psv = ParagraphSXSView()
        psv.footnote_tpl = Template("[{{footnote.reference}}]")
        sxs = {
            'paragraphs': [
                'This is paragraph 1, I mean zero',
                'Will the real paragraph one please stand up?',
                'I am paragraph 10. Is that good enough?'],
            'footnote_refs': [
                {'reference': '1', 'paragraph': 0, 'offset': 7},
                {'reference': '2', 'paragraph': 0, 'offset': 27},
                {'reference': '12', 'paragraph': 2, 'offset': 1}],
            'children': [
                {'paragraphs': ['Subparagraph here'],
                 'footnote_refs': [
                     {'reference': '22', 'paragraph': 0, 'offset': 12}],
                 'children': []}]}
        psv.footnote_refs(sxs)
        self.assertEqual('This is[1] paragraph 1, I mean[2] zero',
                         sxs['paragraphs'][0])
        self.assertEqual('Will the real paragraph one please stand up?',
                         sxs['paragraphs'][1])
        self.assertEqual('I[12] am paragraph 10. Is that good enough?',
                         sxs['paragraphs'][2])
        self.assertEqual('Subparagraph[22] here',
                         sxs['children'][0]['paragraphs'][0])

    @patch('regulations.views.partial_sxs.generator')
    @patch('regulations.views.partial_sxs.notices')
    @patch.object(ParagraphSXSView, 'further_analyses')
    def test_get_context_data(self, further_analyses, notices, generator):
        psv = ParagraphSXSView()
        generator.get_sxs.return_value = {
            'labels': ['lablab', 'another-label'],
            'children': [],
            'page': 1234,
            'paragraphs': ['some', 'content']
        }
        context = psv.get_context_data(label_id='lablab', notice_id='nnnn',
                                       version='vvv')
        self.assertTrue('sxs' in context)
        self.assertTrue('label' in context['sxs'])



from unittest import TestCase

from django.test import RequestFactory

from regulations.generator.sidebar import help


class HelpSidebarTests(TestCase):
    def setUp(self):
        self.req_factory = RequestFactory()
        self.sidebar = help.Help('111-11', 'vvvv')

    def test_context_default_subtemplates(self):
        """No request parameters should include multiple subtemplates"""
        result = self.sidebar.context(None, self.req_factory.get('/'))
        self.assertTrue(result['subtemplates'])

    def test_context_requested_subtemplates(self):
        """Specific request parameters should include only subtemplates which
        exist"""
        result = self.sidebar.context(
            None, self.req_factory.get('/?layers=internal,meta,fake,terms'))
        self.assertEqual(2, len(result['subtemplates']))



# vim: set fileencoding=utf-8
from unittest import TestCase

from regulations.generator.layers.toc_applier import (
    TableOfContentsLayer)


class TableOfContentsLayerTest(TestCase):

    def test_section(self):
        toc = TableOfContentsLayer(None)
        el = {}
        toc.section(el, {'index': ['1']})
        self.assertEqual({}, el)

        toc.section(el, {'index': ['1', '2', '3']})
        self.assertEqual({}, el)

        toc.section(el, {'index': ['1', 'B']})
        self.assertEqual({}, el)

        toc.section(el, {'index': ['1', 'Interpretations']})
        self.assertEqual({}, el)

        toc.section(el, {'index': ['1', '2'], 'title': u'§ 1.2 - Awesome'})
        self.assertEqual(el, {
            'is_section': True,
            'is_section_span': False,
            'section_id': '1-2',
            'label': '1.2',
            'sub_label': 'Awesome'
        })

        toc.section(el, {'index': ['2', '1'], 'title': u'§ 2.1 Sauce'})
        self.assertEqual(el, {
            'is_section': True,
            'is_section_span': False,
            'section_id': '2-1',
            'label': '2.1',
            'sub_label': 'Sauce'
        })

    def test_section_with_thin_space(self):
        toc = TableOfContentsLayer(None)
        el = {}
        toc.section(el, {'index': ['1', '2'], 'title': u'§ 1.2 -  Awesome'})
        self.assertEqual(el, {
            'is_section': True,
            'is_section_span': False,
            'section_id': '1-2',
            'label': '1.2',
            'sub_label': 'Awesome'
        })

    def test_section_span(self):
        toc = TableOfContentsLayer(None)
        el = {}

        toc.section(el, {
            'index': ['1', '2'],
            'title': u'§§ 1.2-6 - This is a span'
        })
        self.assertEqual(el, {
            'is_section': True,
            'is_section_span': True,
            'section_id': '1-2',
            'label': '1.2-6',
            'sub_label': 'This is a span'
        })

    def test_appendix_supplement(self):
        toc = TableOfContentsLayer(None)
        el = {}
        toc.appendix_supplement(el, {'index': ['1']})
        self.assertEqual({}, el)

        toc.appendix_supplement(el, {'index': ['1', '2', '3']})
        self.assertEqual({}, el)

        toc.appendix_supplement(el, {'index': ['1', 'B', '3']})
        self.assertEqual({}, el)

        toc.appendix_supplement(el, {'index': ['1', 'Interp', '3']})
        self.assertEqual({}, el)

        toc.appendix_supplement(el, {
            'index': ['1', 'B'],
            'title': 'Appendix B - Bologna'})
        self.assertEqual(el, {
            'is_appendix': True,
            'is_first_appendix': True,
            'label': 'Appendix B',
            'sub_label': 'Bologna',
            'section_id': '1-B'
        })

        el = {}
        toc.appendix_supplement(el, {
            'index': ['204', 'A'],
            'title': 'Appendix A to Part 204 - Model Forms'})
        self.assertEqual(el, {
            'is_appendix': True,
            'is_first_appendix': True,
            'label': 'Appendix A to Part 204',
            'sub_label': 'Model Forms',
            'section_id': '204-A'
        })

        el = {}
        toc.appendix_supplement(el, {
            'index': ['1', 'Interp'],
            'title': 'Supplement I to 8787 - I am Iron Man'})
        self.assertEqual(el, {
            'is_supplement': True,
            'label': 'Supplement I to 8787',
            'sub_label': 'I am Iron Man',
            'section_id': '1-Interp'
        })

    def test_apply_layer_url(self):
        toc = TableOfContentsLayer({'100': [
            {'title': u'§ 100.1 Intro', 'index': ['100', '1']}]})

        result = toc.apply_layer('100')
        self.assertEqual('#100-1', result[1][0]['url'])

        toc.sectional = True
        toc.version = 'verver'
        result = toc.apply_layer('100')
        self.assertTrue('100-1/verver#100-1' in result[1][0]['url'])

    def test_apply_layer_compatibility(self):
        toc = TableOfContentsLayer({'100': [
            {'title': u'§ 100.1 Intro', 'index': ['100', '1']},
            {'title': 'Appendix A', 'index': ['100', 'A']},
            {'title': 'Supplement I', 'index': ['100', 'Interp']}]})
        _, result = toc.apply_layer('100')
        self.assertEqual(3, len(result))

        toc = TableOfContentsLayer({
            '100': [
                {'title': 'Subpart A', 'index': ['100', 'Subpart', 'A']},
                {'title': 'Appendix A', 'index': ['100', 'A']},
                {'title': 'Supplement I', 'index': ['100', 'Interp']}],
            '100-Subpart-A': [
                {'title': u'§ 100.1 Intro', 'index': ['100', '1']},
                {'title': u'§ 100.2 Sec2', 'index': ['100', '2']},
                {'title': u'§ 100.3 Sec3', 'index': ['100', '3']}]
            })
        _, result = toc.apply_layer('100')
        self.assertEqual(3, len(result))
        self.assertEqual(3, len(result[0]['sub_toc']))

    def test_apply_layer_first_appendix(self):
        toc = TableOfContentsLayer({'100': [
            {'title': 'Appendix A', 'index': ['100', 'A']},
            {'title': 'Appendix B', 'index': ['100', 'B']},
            {'title': 'Appendix C', 'index': ['100', 'C']},
            {'title': 'Supplement I', 'index': ['100', 'Interp']}]})
        _, result = toc.apply_layer('100')
        self.assertEqual(4, len(result))
        aA, aB, aC, sI = result
        self.assertTrue(aA['is_first_appendix'])
        self.assertFalse(aB['is_first_appendix'])
        self.assertFalse(aC['is_first_appendix'])
        self.assertFalse(sI.get('is_first_appendix', False))

        toc = TableOfContentsLayer({'100': [
            {'title': 'Supplement I', 'index': ['100', 'Interp']}]})
        _, result = toc.apply_layer('100')
        self.assertEqual(1, len(result))
        self.assertFalse(result[0].get('is_first_appendix', False))

    def test_apply_layer_interp_emptysubpart(self):
        toc = TableOfContentsLayer({'100': [
            {'title': u'§ 100.1 Intro', 'index': ['100', '1']},
            {'title': u'§ 100.2 Second', 'index': ['100', '2']},
            {'title': 'Supplement I', 'index': ['100', 'Interp']}]})
        _, result = toc.apply_layer('100')
        self.assertEqual(3, len(result))
        s1, s2, interp = result
        self.assertEqual(1, len(interp['sub_toc']))
        nosubpart = interp['sub_toc'][0]
        self.assertEqual('Regulation Text', nosubpart['label'])
        self.assertEqual(['100', 'Subpart', 'Interp'], nosubpart['index'])

        toc = TableOfContentsLayer({'100': [
            {'title': u'§ 100.1 Intro', 'index': ['100', '1']},
            {'title': u'§ 100.2 Second', 'index': ['100', '2']},
            {'title': 'Appendix A', 'index': ['100', 'A']},
            {'title': 'Appendix C', 'index': ['100', 'C']},
            {'title': 'Supplement I', 'index': ['100', 'Interp']}]})
        _, result = toc.apply_layer('100')
        self.assertEqual(5, len(result))
        s1, s2, appA, appC, interp = result
        self.assertEqual(2, len(interp['sub_toc']))
        nosubpart, appendices = interp['sub_toc']
        self.assertEqual('Regulation Text', nosubpart['label'])
        self.assertEqual(['100', 'Subpart', 'Interp'], nosubpart['index'])
        self.assertEqual('Appendices', appendices['label'])
        self.assertEqual(['100', 'Appendices', 'Interp'], appendices['index'])

    def test_apply_layer_interp_subparts(self):
        toc = TableOfContentsLayer({
            '100': [
                {'title': 'Subpart A', 'index': ['100', 'Subpart', 'A']},
                {'title': 'Supplement I', 'index': ['100', 'Interp']}],
            '100-Subpart-A': [
                {'title': u'§ 100.1 Intro', 'index': ['100', '1']},
                {'title': u'§ 100.2 Second', 'index': ['100', '2']}]})
        _, result = toc.apply_layer('100')
        self.assertEqual(2, len(result))
        subpartA, interp = result
        self.assertEqual(2, len(subpartA['sub_toc']))
        self.assertEqual(1, len(interp['sub_toc']))
        nosubpart = interp['sub_toc'][0]
        self.assertEqual('Subpart A', nosubpart['label'])
        self.assertEqual(['100', 'Subpart', 'A', 'Interp'], nosubpart['index'])

        toc = TableOfContentsLayer({
            '100': [
                {'title': 'Subpart A', 'index': ['100', 'Subpart', 'A']},
                {'title': 'Appendix A', 'index': ['100', 'A']},
                {'title': 'Appendix C', 'index': ['100', 'C']},
                {'title': 'Supplement I', 'index': ['100', 'Interp']}],
            '100-Subpart-A': [
                {'title': u'§ 100.1 Intro', 'index': ['100', '1']},
                {'title': u'§ 100.2 Second', 'index': ['100', '2']}]})
        _, result = toc.apply_layer('100')
        self.assertEqual(4, len(result))
        subpartA, appA, appC, interp = result
        self.assertEqual(2, len(interp['sub_toc']))
        nosubpart, appendices = interp['sub_toc']
        self.assertEqual('Subpart A', nosubpart['label'])
        self.assertEqual(['100', 'Subpart', 'A', 'Interp'], nosubpart['index'])
        self.assertEqual('Appendices', appendices['label'])
        self.assertEqual(['100', 'Appendices', 'Interp'], appendices['index'])



from unittest import TestCase

from six.moves.urllib.parse import parse_qs
from django.template import RequestContext
from django.template.loader import get_template
from django.test import RequestFactory, SimpleTestCase, override_settings


class TemplateTest(TestCase):
    def test_title_in_base(self):
        context = {
            'env': 'dev',
            'reg_part': '204',
            'meta': {'cfr_title_number': '2'}}
        request = RequestFactory().get('/fake-path')
        c = RequestContext(request, context)
        t = get_template('regulations/base.html')
        rendered = t.render(c)

        title = '2 CFR Part 204 | eRegulations'
        self.assertTrue(title in rendered)


class GlobalContextTest(SimpleTestCase):

    @override_settings(JS_DEBUG=True)
    def test_debug(self):
        resp = self.client.get('/about')
        self.assertEqual(resp.context['EREGS_GLOBALS']['ENV'], 'source')

    @override_settings(JS_DEBUG=False)
    def test_prod(self):
        resp = self.client.get('/about')
        self.assertEqual(resp.context['EREGS_GLOBALS']['ENV'], 'built')

    @override_settings(
        ANALYTICS={
            'GOOGLE': {
                'GTM_SITE_ID': 'gtm-site-id',
                'GA_SITE_ID': 'ga-site-id',
            },
            'DAP': {
                'AGENCY': 'agency',
                'SUBAGENCY': 'sub-agency',
            },
        },
    )
    def test_analytics(self):
        resp = self.client.get('/about')
        analytics = resp.context['EREGS_GLOBALS']['ANALYTICS']
        self.assertEquals('gtm-site-id',
                          analytics['GOOGLE']['GTM_SITE_ID'])
        self.assertEquals('ga-site-id',
                          analytics['GOOGLE']['GA_SITE_ID'])
        self.assertEquals('agency', analytics['DAP']['AGENCY'])
        self.assertEquals('sub-agency',
                          analytics['DAP']['SUBAGENCY'])
        self.assertEquals(
            parse_qs('agency=agency&subagency=sub-agency'),
            parse_qs(analytics['DAP']['DAP_URL_PARAMS']),
        )

    def test_prefix(self):
        resp = self.client.get('/about')
        self.assertIn('APP_PREFIX', resp.context['EREGS_GLOBALS'])



from unittest import TestCase
from regulations.generator.layers import diff_applier
from regulations.generator.layers import tree_builder
from regulations.generator.node_types import REGTEXT

from collections import deque


class DiffApplierTest(TestCase):

    def test_create_applier(self):
        diff = {'some': 'diff'}
        da = diff_applier.DiffApplier(diff, None)
        self.assertEquals(da.diff, diff)

    def create_diff_applier(self):
        diff = {'some': 'diff'}
        da = diff_applier.DiffApplier(diff, None)
        text = "abcd"
        da.deconstruct_text(text)
        return da

    def test_deconstruct_text(self):
        da = self.create_diff_applier()

        self.assertTrue('oq' in da.__dict__)

        deque_list = [
            deque(['a']), deque(['b']),
            deque(['c']), deque(['d'])]

        self.assertEquals(da.oq, deque_list)

    def test_insert_text(self):
        da = self.create_diff_applier()

        da.insert_text(1, 'AA')
        deque_list = [
            deque(['a']), deque(['<ins>AA</ins>', 'b']),
            deque(['c']), deque(['d'])]
        self.assertEquals(da.oq, deque_list)

    def test_insert_text_at_end(self):
        da = self.create_diff_applier()

        da.insert_text(4, 'AA')
        deque_list = [
            deque(['a']), deque(['b']),
            deque(['c']), deque(['d', '<ins>', 'AA', '</ins>'])]
        self.assertEquals(da.oq, deque_list)

    def test_delete_text(self):
        da = self.create_diff_applier()
        da.delete_text(0, 2)
        deque_list = [
            deque(['<del>', 'a']), deque(['b', '</del>']),
            deque(['c']), deque(['d'])]
        self.assertEquals(da.oq, deque_list)

    def test_apply_diff(self):
        diff = {'204': {'text': [('delete', 0, 2), ('insert', 4, 'AAB')],
                        'op': ''}}
        da = diff_applier.DiffApplier(diff, None)
        da.apply_diff('acbd', '204')
        new_text = da.get_text()
        self.assertEquals('<del>ac</del>bd<ins>AAB</ins>', new_text)

    def test_apply_diff_title(self):
        diff = {'204': {'title': [('delete', 0, 2), ('insert', 4, 'AAC')],
                        'text':  [('delete', 0, 2), ('insert', 4, 'AAB')],
                        'op': ''}}
        da = diff_applier.DiffApplier(diff, None)
        da.apply_diff('acbd', '204', component='title')
        new_text = da.get_text()
        self.assertEquals('<del>ac</del>bd<ins>AAC</ins>', new_text)

    def test_delete_all(self):
        da = diff_applier.DiffApplier({}, None)
        original = 'abcd'
        deleted = da.delete_all(original)
        self.assertEqual('<del>abcd</del>', deleted)

    def build_tree(self):
        child = {
            'text': 'child text',
            'children': [],
            'label_id': '204-3',
            'label': ['204', '3'],
            'node_type': REGTEXT
        }
        tree = {
            'text': 'parent text',
            'children': [child],
            'label_id': '204',
            'label': ['204'],
            'node_type': REGTEXT
        }
        return tree

    def test_add_nodes_to_tree(self):
        tree = self.build_tree()

        new_node = {
            'text': 'child text',
            'children': [],
            'label_id': '204-2',
            'label': ['204', '2'],
            'node_type': REGTEXT
        }

        da = diff_applier.DiffApplier({}, None)
        adds = tree_builder.AddQueue()
        adds.insert_all([('204-2', new_node)])

        da.add_nodes_to_tree(tree, adds)
        self.assertEqual(len(tree['children']), 2)
        self.assertEqual(tree['children'][0], new_node)

    def test_add_nodes_new_section(self):
        tree = self.build_tree()
        new_node = {
            'text': 'new node text',
            'children': [],
            'label_id': '204-2',
            'label': ['204', '2'],
            'node_type': REGTEXT
        }
        new_node_child = {
            'text': 'new node child text',
            'children': [],
            'label_id': '204-2-a',
            'label': ['204', '2', 'a'],
            'node_type': REGTEXT
        }
        da = diff_applier.DiffApplier({}, None)
        adds = tree_builder.AddQueue()
        adds.insert_all([('204-2-a', new_node_child), ('204-2', new_node)])
        da.add_nodes_to_tree(tree, adds)
        self.assertEqual(len(tree['children']), 2)

        new_node['children'].append(new_node_child)
        self.assertEqual(tree['children'][0], new_node)

    def test_add_nodes_empty_tree(self):
        tree = {}
        new_node = {
            'text': 'new node text',
            'children': [],
            'label_id': '204-2',
            'label': ['204', '2'],
            'node_type': REGTEXT
        }
        da = diff_applier.DiffApplier({}, None)
        adds = tree_builder.AddQueue()
        adds.insert_all([('204-2', new_node)])

        da.add_nodes_to_tree(tree, adds)

    def test_add_nodes_child_ops(self):
        """If we don't know the correct order of children, attempt to use data
        from `child_ops`"""
        def mk_node(section):
            return {'label': ['100', section], 'node_type': 'regtext',
                    'children': []}
        node_1, node_2 = mk_node('1'), mk_node('2')
        node_2a, node_3 = mk_node('2a'), mk_node('3')
        original = {'label': ['100'], 'node_type': 'regtext', 'children': [
            node_1, node_2
        ]}

        diff = {
            '100': {
                'op': diff_applier.DiffApplier.MODIFIED_OP,
                'child_ops': [[diff_applier.DiffApplier.EQUAL, 0, 1],
                              [diff_applier.DiffApplier.DELETE, 1, 2],
                              [diff_applier.DiffApplier.INSERT, 1, ['100-2a']],
                              [diff_applier.DiffApplier.INSERT, 2, ['100-3']]]
            },
            '100-2': {'op': diff_applier.DiffApplier.DELETED_OP},
            '100-2a': {
                'op': diff_applier.DiffApplier.ADDED_OP,
                'node': node_2a
            },
            '100-3': {
                'op': diff_applier.DiffApplier.ADDED_OP,
                'node': node_3
            }
        }

        adds = tree_builder.AddQueue()
        adds.insert_all([('100-2a', node_2a), ('100-3', node_3)])

        da = diff_applier.DiffApplier(diff, None)
        da.add_nodes_to_tree(original, adds)
        self.assertEqual(original['child_labels'],
                         ['100-1', '100-2', '100-2a', '100-3'])

    def test_set_child_labels_reorder(self):
        """Nodes which have been _moved_ should be ordered in their final
        resting position"""
        node1 = {'label': ['1', '1'], 'node_type': 'regtext', 'children': []}
        node2 = {'label': ['1', '2'], 'node_type': 'regtext', 'children': []}
        root = {'label': ['1'], 'node_type': 'regtext',
                'children': [node1, node2]}

        diff = {'1': {
            'op': diff_applier.DiffApplier.MODIFIED_OP,
            'child_ops': [[diff_applier.DiffApplier.DELETE, 0, 2],
                          [diff_applier.DiffApplier.INSERT, 0, ['1-2', '1-1']]]
        }}

        da = diff_applier.DiffApplier(diff, None)
        da.set_child_labels(root)
        self.assertEqual(root['child_labels'], ['1-2', '1-1'])

    def test_child_picking(self):
        da = self.create_diff_applier()
        da.label_requested = '204-3'

        self.assertTrue(da.is_child_of_requested('204-3-a'))
        self.assertFalse(da.is_child_of_requested('204-32'))

        da.label_requested = '204-3-Interp'
        self.assertTrue(da.is_child_of_requested('204-3-a-Interp'))
        self.assertTrue(da.is_child_of_requested('204-3-Interp-1'))
        self.assertFalse(da.is_child_of_requested('204-30-Interp'))

        da.label_requested = '204-3-Interp-4'
        self.assertFalse(da.is_child_of_requested('204-3-a-Interp'))
        self.assertFalse(da.is_child_of_requested('204-3-Interp-1'))
        self.assertTrue(da.is_child_of_requested('204-3-Interp-4-a'))

    def test_tree_changes_new_section(self):
        diff = {'9999-25': {'op': 'added',
                            'node': {'text': 'Some Text',
                                     'node_type': 'REGTEXT',
                                     'label': ['9999', '25'],
                                     'children': []}}}
        old_tree = {}
        da = diff_applier.DiffApplier(diff, '9999-25')
        da.tree_changes(old_tree)

        self.assertEqual(old_tree['label'], ['9999', '25'])



# vim: set encoding=utf-8
from unittest import TestCase
from mock import patch

from django.conf import settings
from django.test import RequestFactory

from regulations.views import utils


class UtilsTest(TestCase):
    def setUp(self):

        if hasattr(settings, 'ANALYTICS'):
            self.old_analytics = settings.ANALYTICS
        if hasattr(settings, 'JS_DEBUG'):
            self.old_js_debug = settings.JS_DEBUG

    def tearDown(self):
        if hasattr(self, 'old_analytics'):
            settings.ANALYTICS = self.old_analytics

        if hasattr(self, 'old_js_debug'):
            settings.JS_DEBUG = self.old_js_debug

    def test_get_layer_list(self):
        names = 'meta,meta,GRAPHICS,fakelayer,internal'
        layer_list = utils.get_layer_list(names)
        self.assertEquals(set(['meta', 'internal', 'graphics']), layer_list)

    def test_layer_names(self):
        request = RequestFactory().get('/?layers=graphics,meta,other')
        self.assertEqual(utils.layer_names(request), set(['graphics', 'meta']))

        request = RequestFactory().get('/?layers=')
        self.assertEqual(utils.layer_names(request), set())

        request = RequestFactory().get('/')
        self.assertTrue(len(utils.layer_names(request)) > 4)

    @patch('regulations.views.utils.fetch_toc')
    def test_first_section(self, fetch_toc):
        fetch_toc.return_value = [
            {'section_id': '204-100', 'index': ['204', '100']},
            {'section_id': '204-101', 'index': ['204', '101']}]
        first = utils.first_section('204', '2')
        self.assertEqual(first, '204-100')

    def test_make_sortable(self):
        """Verify that strings get decomposed correctly into sortable tuples"""
        self.assertEqual(utils.make_sortable("abc"), ("abc",))
        self.assertEqual(utils.make_sortable("123"), (123,))
        self.assertEqual(utils.make_sortable("abc123def456"),
                         ("abc", 123, "def", 456))
        self.assertEqual(utils.make_sortable("123abc456"), (123, "abc", 456))

    @patch('regulations.views.utils.api_reader')
    def test_regulation_meta_404(self, api_reader):
        """We shouldn't crash if meta data isn't available"""
        ret_vals = [None, {}, {'111-22': 'something'}]
        for ret_val in ret_vals:
            api_reader.ApiReader.return_value.layer.return_value = ret_val
            self.assertEqual({}, utils.regulation_meta('111', 'vvv'))



from datetime import datetime

from django.http import HttpResponse
from django.template import RequestContext
from django.template.loader import select_template
from regulations.views import utils

from regulations.generator import versions


def filter_future_amendments(versions):
    """ Take a list of amendments, and only return a list of those that are in
    the future. """
    today = datetime.today()
    amendments = [v for v in versions if v['by_date'] > today]
    amendments.sort(key=lambda v: v['by_date'])
    return amendments


def get_regulations_list(all_versions):
    """ Given a list of regulation versions, add data about those regulations
    to that list. """

    regs = []
    reg_parts = sorted(all_versions.keys(), key=utils.make_sortable)

    for part in reg_parts:
        version = all_versions[part][0]['version']
        reg_meta = utils.regulation_meta(part, version)
        first_section = utils.first_section(part, version)
        amendments = filter_future_amendments(all_versions.get(part, None))

        reg = {'part': part,
               'meta': reg_meta,
               'reg_first_section': first_section,
               'amendments': amendments}

        regs.append(reg)
    return regs


def universal(request):
    context = {}

    all_versions = versions.fetch_regulations_and_future_versions()
    regs = get_regulations_list(all_versions)

    context['regulations'] = regs
    context['cfr_title_text'] = regs[0]['meta']['cfr_title_text']
    context['cfr_title_number'] = utils.to_roman(
        regs[0]['meta']['cfr_title_number'])
    context['cfr_titleno_arabic'] = regs[0]['meta']['cfr_title_number']

    c = RequestContext(request, context)
    t = select_template([
        'regulations/universal_landing.html',
        'regulations/generic_universal.html'])
    return HttpResponse(t.render(c))



import abc

from regulations.generator.subterp import filter_by_subterp


class SidebarBase(object):
    """Base class for Sidebar components. Provides an interface"""
    __metaclass__ = abc.ABCMeta

    def __init__(self, label_id, version):
        self.label_id = label_id
        self.label_parts = label_id.split('-')
        self.cfr_part = self.label_parts[0]
        self.version = version

    @abc.abstractproperty
    def shorthand(self):
        """Used for deriving templates, etc."""
        raise NotImplementedError

    @abc.abstractmethod
    def context(self, http_client, request):
        """Create a context dictionary specific to this sidebar type"""
        raise NotImplementedError

    def full_context(self, http_client, request):
        subcontext = self.context(http_client, request)
        subcontext.update(
            template_name='regulations/sidebar/{}.html'.format(self.shorthand)
        )
        return subcontext

    def fetch_relevant_trees(self, http_client):
        """If using subterps, we might be getting a list of relevant trees
        rather than a single node."""
        is_interp = 'Interp' in self.label_parts
        is_complex = set(['Subpart', 'Appendices']) & set(self.label_parts)
        is_subterp = is_interp and is_complex
        if is_subterp:
            interp = http_client.regulation(
                self.cfr_part + '-Interp', self.version)
            if interp:
                for tree_node in filter_by_subterp(
                        interp['children'], self.label_parts, self.version):
                    yield tree_node
        else:
            node = http_client.regulation(self.label_id, self.version)
            if node:
                yield node



import boto3
from django.conf import settings
from django.core.management.base import BaseCommand


class Command(BaseCommand):
    help = 'Set CORS rules on the Notice and Comment attachment bucket'

    def handle(self, *args, **options):
        hosts = settings.ALLOWED_HOSTS
        origins = ['http://' + host for host in hosts]
        origins = origins + ['https://' + host for host in hosts]

        session = boto3.Session(
            aws_access_key_id=settings.ATTACHMENT_ACCESS_KEY_ID,
            aws_secret_access_key=settings.ATTACHMENT_SECRET_ACCESS_KEY,
        )
        s3 = session.client('s3')

        s3.put_bucket_cors(
            Bucket=settings.ATTACHMENT_BUCKET,
            CORSConfiguration={
                'CORSRules': [
                    {
                        'AllowedMethods': ['GET', 'PUT'],
                        'AllowedOrigins': origins or ['*'],
                        'AllowedHeaders': ['*'],
                    },
                ],
            },
        )



import re

from django.shortcuts import redirect

from regulations.generator.api_reader import ApiReader
from regulations.generator.versions import fetch_grouped_history
from regulations.views.error_handling import handle_generic_404


def redirect_by_date(request, label_id, year, month, day):
    """If a user requests a date as the version, find the version which was
    current as of that date"""
    date_versions = []
    client = ApiReader()
    for struct in client.regversions(label_id.split('-')[0])['versions']:
        if 'by_date' in struct:
            date_versions.append((struct['by_date'], struct['version']))

    date_versions = sorted(date_versions)
    last_version = None
    date_str = '%s-%s-%s' % (year, month, day)
    while date_versions and date_versions[0][0] <= date_str:
        last_version = date_versions[0][1]
        date_versions = date_versions[1:]

    label_parts = label_id.split('-')
    if last_version and len(label_parts) == 2:
        return redirect('chrome_section_view', label_id, last_version)
    elif last_version and label_parts[-1] == 'Interp':
        return redirect('chrome_section_view', label_id, last_version)
    elif last_version and len(label_parts) == 1:
        return redirect('chrome_regulation_view', label_id, last_version)
    elif last_version:
        return redirect('chrome_paragraph_view', label_id, last_version)
    else:
        return handle_generic_404(request)


def redirect_by_date_get(request, label_id):
    """Handles date, etc. if they are part of the GET variable. We check for
    bad data here (as we can't rely on url regex)"""
    try:
        year = abs(int(request.GET.get('year', '')))
        month = abs(int(request.GET.get('month', '')))
        day = abs(int(request.GET.get('day', '')))

        if year < 100:  # Assume two-digit years are for 2000
            year = 2000 + year

        return redirect_by_date(request, label_id, "%04d" % year,
                                "%02d" % month, "%02d" % day)
    except ValueError:
        return handle_generic_404(request)


def order_diff_versions(label_id, version, new_version):
    # Re-order if needed - History is sorted in reverse chronological order
    for major_version in fetch_grouped_history(label_id.split('-')[0]):
        for notice in major_version['notices']:
            # Hit the "old" version first, meaning it's not actually the old
            # version
            if notice['document_number'] == version:
                return redirect('chrome_section_diff_view', label_id,
                                new_version, version)
            # Hit the new version first -- sort is correct
            elif notice['document_number'] == new_version:
                return redirect('chrome_section_diff_view', label_id,
                                version, new_version)

    # Didn't find the versions in question. Assume this was intentional
    return redirect('chrome_section_diff_view', label_id, version,
                    new_version)


def diff_redirect(request, label_id, version):
    """Handles constructing the diff url by pulling the new version from
    GET. We check for bad data here (as we can't rely on url regex)"""
    new_version = request.GET.get('new_version', '')
    if not re.match(r'^[-\d\w]+$', new_version):
        return handle_generic_404(request)

    response = order_diff_versions(label_id, version, new_version)
    response['Location'] += '?from_version=%s' % version
    return response



import json
import os
from django.conf import settings
import requests


class ApiClient:
    """ Actually go out and make the GET request, or read the disk to acquire
    the required data. """

    def __init__(self):
        self.base_url = settings.API_BASE

    def get_from_file_system(self, suffix):
        if os.path.isdir(self.base_url + suffix):
            suffix = suffix + "/index.html"
        f = open(self.base_url + suffix)
        content = f.read()
        f.close()
        return json.loads(content)

    def get(self, suffix, params={}):
        """Make the GET request. Assume the result is JSON. Right now, there is
        no error handling"""

        if self.base_url.startswith('http'):
            r = requests.get(self.base_url + suffix, params=params)
            if r.status_code == requests.codes.ok:
                return r.json()
            elif r.status_code == 404:
                return None
            else:
                r.raise_for_status()
        else:
            return self.get_from_file_system(suffix)



# vim: set fileencoding=utf-8
from unittest import TestCase
from regulations.generator.link_flattener import flatten_links


class LinkFlattenerTest(TestCase):
    def test_no_links(self):
        self.assertEqual("foo", flatten_links("foo"))

    def test_single_link(self):
        self.assertEqual(
            "<a href=foo> Fee Fie Foe </a>",
            flatten_links("<a href=foo> Fee Fie Foe </a>"))

    def test_unembedded_links(self):
        self.assertEqual(
            "<a href=foo> Fee </a> Fie <a href=bar> Foe </a>",
            flatten_links("<a href=foo> Fee </a> Fie <a href=bar> Foe </a>"))

    def test_embedded_link(self):
        self.assertEqual(
            "<a href=foo> Fee  Fie  Foe </a>",
            flatten_links("<a href=foo> Fee <a href=bar> Fie </a> Foe </a>"))

    def test_multiple_serial_embedded_links(self):
        self.assertEqual(
            "<a href=foo>FeeFieFoe</a>",
            flatten_links(
                "<a href=foo>Fee<a href=bar>Fie</a><a href=baz>Foe</a></a>")
        )

    def test_multiple_level_embedded_links(self):
        self.assertEqual(
            '<a href="1">A B C D E</a>',
            flatten_links(
                '<a href="1">A <a href="2">B <a href="3">C</a> D</a> E</a>')
        )

    def test_real_world_example(self):
        original = """
<span class="stripped-marker">a.</span><span class="paragraph-marker">(a)
</span> All of those items on <a href="/447-11/2015-12992#447-11-p3367907674"
class="citation definition" data-definition="447-11-p3367907674"
data-defined-term="the u.s. munitions import list">the U.S. Munitions Import
List</a> (see \xa7&nbsp;<a href="/447-21/2015-12992#447-21" class="citation
internal" data-section-id="447-21">447.21</a>) which are \u201c
<a href="/447-11/2015-12992#447-11-p21112113" class="citation definition"
data-definition="447-11-p21112113" data-defined-term="firearms">firearms</a>
\u201d or \u201cammunition\u201d as defined in 18
<a href="/447-21/2015-12992#447-21-p4-p67-c" class="citation definition"
data-definition="447-21-p4-p67-c" data-defined-term="u">U</a>.S.
<a href="/447-21/2015-12992#447-21-p4-p67-a" class="citation definition"
data-definition="447-21-p4-p67-a" data-defined-term="c">C</a>. 921(a) are
subject to the interstate and foreign commerce controls contained in Chapter
44 of Title 18 <a href="/447-21/2015-12992#447-21-p4-p67-c"
class="citation definition" data-definition="447-21-p4-p67-c"
data-defined-term="u">U</a>.S.<a href="/447-21/2015-12992#447-21-p4-p67-a"
class="citation definition" data-definition="447-21-p4-p67-a"
data-defined-term="c">C</a>. and <a href="http://api.fdsys.gov/link?
titlenum=27&partnum=478&collection=cfr&year=mostrecent"
class="citation external" target="_blank">27
<a href="/447-11/2015-12992#447-11-p3520592195" class="citation definition"
data-definition="447-11-p3520592195" data-defined-term="cfr">CFR</a> Part 478
</a> and if they are \u201c<a href="/447-11/2015-12992#447-11-p21112113"
class="citation definition" data-definition="447-11-p21112113"
data-defined-term="firearms">firearms</a>\u201d within the definition set out
in 26 <a href="/447-21/2015-12992#447-21-p4-p67-c" class="citation definition"
data-definition="447-21-p4-p67-c" data-defined-term="u">U</a>.S.
<a href="/447-21/2015-12992#447-21-p4-p67-a" class="citation definition"
data-definition="447-21-p4-p67-a" data-defined-term="c">C</a>. 5845(a) are
also subject to the provisions of <a href="http://api.fdsys.gov/link?
titlenum=27&partnum=479&collection=cfr&year=mostrecent"
class="citation external" target="_blank">27
<a href="/447-11/2015-12992#447-11-p3520592195" class="citation definition"
data-definition="447-11-p3520592195" data-defined-term="cfr">CFR</a> Part 479
</a>. Any <a href="/447-11/2015-12992#447-11-p3550103376"
class="citation definition" data-definition="447-11-p3550103376"
data-defined-term="person">person</a> engaged in the business of importing
<a href="/447-11/2015-12992#447-11-p21112113" class="citation definition"
data-definition="447-11-p21112113" data-defined-term="firearms">firearms</a>
or ammunition as defined in 18 <a href="/447-21/2015-12992#447-21-p4-p67-c"
class="citation definition" data-definition="447-21-p4-p67-c"
data-defined-term="u">U</a>.S.<a href="/447-21/2015-12992#447-21-p4-p67-a"
class="citation definition" data-definition="447-21-p4-p67-a"
data-defined-term="c">C</a>. 921(a) must obtain a
<a href="/447-11/2015-12992#447-11-p3473882945" class="citation definition"
data-definition="447-11-p3473882945" data-defined-term="license">license</a>
under the provisions of <a href="http://api.fdsys.gov/link?
titlenum=27&partnum=478&collection=cfr&year=mostrecent"
class="citation external" target="_blank">27
<a href="/447-11/2015-12992#447-11-p3520592195" class="citation definition"
data-definition="447-11-p3520592195" data-defined-term="cfr">CFR</a> Part 478
</a>, and if he <a href="/447-11/2015-12992#447-11-p3984137023"
class="citation definition" data-definition="447-11-p3984137023"
data-defined-term="import">imports</a>
<a href="/447-11/2015-12992#447-11-p21112113" class="citation definition"
data-definition="447-11-p21112113" data-defined-term="firearms">firearms</a>
which fall within the definition of 26
<a href="/447-21/2015-12992#447-21-p4-p67-c" class="citation definition"
data-definition="447-21-p4-p67-c" data-defined-term="u">U</a>.S.
<a href="/447-21/2015-12992#447-21-p4-p67-a" class="citation definition"
data-definition="447-21-p4-p67-a" data-defined-term="c">C</a>. 5845(a) must
also register and pay special tax pursuant to the provisions of
<a href="http://api.fdsys.gov/link?titlenum=27&partnum=479&collection=cfr
&year=mostrecent" class="citation external" target="_blank">27
<a href="/447-11/2015-12992#447-11-p3520592195" class="citation definition"
data-definition="447-11-p3520592195" data-defined-term="cfr">CFR</a> Part 479
</a>. Such licensing, registration and special tax requirements are in
addition to registration under subpart D of this part.
"""
        expected = """
<span class="stripped-marker">a.</span><span class="paragraph-marker">(a)
</span> All of those items on <a href="/447-11/2015-12992#447-11-p3367907674"
class="citation definition" data-definition="447-11-p3367907674"
data-defined-term="the u.s. munitions import list">the U.S. Munitions Import
List</a> (see \xa7&nbsp;<a href="/447-21/2015-12992#447-21" class="citation
internal" data-section-id="447-21">447.21</a>) which are \u201c
<a href="/447-11/2015-12992#447-11-p21112113" class="citation definition"
data-definition="447-11-p21112113" data-defined-term="firearms">firearms</a>
\u201d or \u201cammunition\u201d as defined in 18
<a href="/447-21/2015-12992#447-21-p4-p67-c" class="citation definition"
data-definition="447-21-p4-p67-c" data-defined-term="u">U</a>.S.
<a href="/447-21/2015-12992#447-21-p4-p67-a" class="citation definition"
data-definition="447-21-p4-p67-a" data-defined-term="c">C</a>. 921(a) are
subject to the interstate and foreign commerce controls contained in Chapter
44 of Title 18 <a href="/447-21/2015-12992#447-21-p4-p67-c"
class="citation definition" data-definition="447-21-p4-p67-c"
data-defined-term="u">U</a>.S.<a href="/447-21/2015-12992#447-21-p4-p67-a"
class="citation definition" data-definition="447-21-p4-p67-a"
data-defined-term="c">C</a>. and <a href="http://api.fdsys.gov/link?
titlenum=27&partnum=478&collection=cfr&year=mostrecent"
class="citation external" target="_blank">27
CFR Part 478
</a> and if they are \u201c<a href="/447-11/2015-12992#447-11-p21112113"
class="citation definition" data-definition="447-11-p21112113"
data-defined-term="firearms">firearms</a>\u201d within the definition set out
in 26 <a href="/447-21/2015-12992#447-21-p4-p67-c" class="citation definition"
data-definition="447-21-p4-p67-c" data-defined-term="u">U</a>.S.
<a href="/447-21/2015-12992#447-21-p4-p67-a" class="citation definition"
data-definition="447-21-p4-p67-a" data-defined-term="c">C</a>. 5845(a) are
also subject to the provisions of <a href="http://api.fdsys.gov/link?
titlenum=27&partnum=479&collection=cfr&year=mostrecent"
class="citation external" target="_blank">27
CFR Part 479
</a>. Any <a href="/447-11/2015-12992#447-11-p3550103376"
class="citation definition" data-definition="447-11-p3550103376"
data-defined-term="person">person</a> engaged in the business of importing
<a href="/447-11/2015-12992#447-11-p21112113" class="citation definition"
data-definition="447-11-p21112113" data-defined-term="firearms">firearms</a>
or ammunition as defined in 18 <a href="/447-21/2015-12992#447-21-p4-p67-c"
class="citation definition" data-definition="447-21-p4-p67-c"
data-defined-term="u">U</a>.S.<a href="/447-21/2015-12992#447-21-p4-p67-a"
class="citation definition" data-definition="447-21-p4-p67-a"
data-defined-term="c">C</a>. 921(a) must obtain a
<a href="/447-11/2015-12992#447-11-p3473882945" class="citation definition"
data-definition="447-11-p3473882945" data-defined-term="license">license</a>
under the provisions of <a href="http://api.fdsys.gov/link?
titlenum=27&partnum=478&collection=cfr&year=mostrecent"
class="citation external" target="_blank">27
CFR Part 478
</a>, and if he <a href="/447-11/2015-12992#447-11-p3984137023"
class="citation definition" data-definition="447-11-p3984137023"
data-defined-term="import">imports</a>
<a href="/447-11/2015-12992#447-11-p21112113" class="citation definition"
data-definition="447-11-p21112113" data-defined-term="firearms">firearms</a>
which fall within the definition of 26
<a href="/447-21/2015-12992#447-21-p4-p67-c" class="citation definition"
data-definition="447-21-p4-p67-c" data-defined-term="u">U</a>.S.
<a href="/447-21/2015-12992#447-21-p4-p67-a" class="citation definition"
data-definition="447-21-p4-p67-a" data-defined-term="c">C</a>. 5845(a) must
also register and pay special tax pursuant to the provisions of
<a href="http://api.fdsys.gov/link?titlenum=27&partnum=479&collection=cfr
&year=mostrecent" class="citation external" target="_blank">27
CFR Part 479
</a>. Such licensing, registration and special tax requirements are in
addition to registration under subpart D of this part.
"""
        self.assertEqual(expected, flatten_links(original))
